# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for Substrate keys handling."""

# Imports
from __future__ import annotations

from functools import lru_cache
from typing import Union

from bip_utils.addr import SubstrateSr25519AddrEncoder
from bip_utils.ecc import IPrivateKey, IPublicKey, Sr25519PrivateKey, Sr25519PublicKey
from bip_utils.substrate.conf import SubstrateCoinConf
from bip_utils.substrate.substrate_ex import SubstrateKeyError
from bip_utils.utils.misc import DataBytes


class SubstratePublicKey:
    """Substrate public key class."""

    m_pub_key: IPublicKey
    m_coin_conf: SubstrateCoinConf

    @classmethod
    def FromBytesOrKeyObject(cls,
                             pub_key: Union[bytes, IPublicKey],
                             coin_conf: SubstrateCoinConf) -> SubstratePublicKey:
        """
        Get the public key from key bytes or object.

        Args:
            pub_key (bytes or IPublicKey)       : Public key
            coin_conf (SubstrateCoinConf object): SubstrateCoinConf object

        Returns:
            SubstratePublicKey object: SubstratePublicKey object

        Raises:
            SubstrateKeyError: If the key constructed from the bytes is not valid
        """
        return (cls.FromBytes(pub_key, coin_conf)
                if isinstance(pub_key, bytes)
                else cls(pub_key, coin_conf))

    @classmethod
    def FromBytes(cls,
                  key_bytes: bytes,
                  coin_conf: SubstrateCoinConf) -> SubstratePublicKey:
        """
        Create from bytes.

        Args:
            key_bytes (bytes)                   : Key bytes
            coin_conf (SubstrateCoinConf object): SubstrateCoinConf object

        Raises:
            SubstrateKeyError: If the key constructed from the bytes is not valid
        """
        return cls(cls.__KeyFromBytes(key_bytes),
                   coin_conf)

    def __init__(self,
                 pub_key: IPublicKey,
                 coin_conf: SubstrateCoinConf) -> None:
        """
        Construct class.

        Args:
            pub_key (IPublicKey object)         : Key object
            coin_conf (SubstrateCoinConf object): SubstrateCoinConf object

        Raises:
            SubstrateKeyError: If the bytes length is not valid
            TypeError: If the key is not a Sr25519PublicKey object
        """
        if not isinstance(pub_key, Sr25519PublicKey):
            raise TypeError("Invalid public key object type")
        self.m_pub_key = pub_key
        self.m_coin_conf = coin_conf

    def KeyObject(self) -> IPublicKey:
        """
        Return the key object.

        Returns:
            IPublicKey object: Key object
        """
        return self.m_pub_key

    @lru_cache()
    def RawCompressed(self) -> DataBytes:
        """
        Return raw compressed public key.

        Returns:
            DataBytes object: DataBytes object
        """
        return self.m_pub_key.RawCompressed()

    @lru_cache()
    def RawUncompressed(self) -> DataBytes:
        """
        Return raw uncompressed public key.

        Returns:
            DataBytes object: DataBytes object
        """
        return self.m_pub_key.RawUncompressed()

    @lru_cache()
    def ToAddress(self) -> str:
        """
        Return the address correspondent to the public key.

        Returns:
            str: Address string
        """
        return SubstrateSr25519AddrEncoder.EncodeKey(self.m_pub_key,
                                                     **self.m_coin_conf.AddrParams())

    @staticmethod
    def __KeyFromBytes(key_bytes: bytes) -> IPublicKey:
        """
        Construct key from bytes.

        Args:
            key_bytes (bytes): Key bytes

        Returns:
            IPublicKey object: IPublicKey object

        Raises:
            SubstrateKeyError: If the key constructed from the bytes is not valid
        """
        try:
            return Sr25519PublicKey.FromBytes(key_bytes)
        except ValueError as ex:
            raise SubstrateKeyError("Invalid public key") from ex


class SubstratePrivateKey:
    """Substrate private key class."""

    m_priv_key: IPrivateKey
    m_coin_conf: SubstrateCoinConf

    @classmethod
    def FromBytesOrKeyObject(cls,
                             priv_key: Union[bytes, IPrivateKey],
                             coin_conf: SubstrateCoinConf) -> SubstratePrivateKey:
        """
        Get the private key from key bytes or object.

        Args:
            priv_key (bytes or IPrivateKey)     : Private key
            coin_conf (SubstrateCoinConf object): SubstrateCoinConf object

        Returns:
            SubstratePrivateKey object: SubstratePrivateKey object

        Raises:
            SubstrateKeyError: If the key constructed from the bytes is not valid
        """
        return (cls.FromBytes(priv_key, coin_conf)
                if isinstance(priv_key, bytes)
                else cls(priv_key, coin_conf))

    @classmethod
    def FromBytes(cls,
                  key_bytes: bytes,
                  coin_conf: SubstrateCoinConf) -> SubstratePrivateKey:
        """
        Create from bytes.

        Args:
            key_bytes (bytes)                   : Key bytes
            coin_conf (SubstrateCoinConf object): SubstrateCoinConf object

        Raises:
            SubstrateKeyError: If the key constructed from the bytes is not valid
        """
        return cls(cls.__KeyFromBytes(key_bytes),
                   coin_conf)

    def __init__(self,
                 priv_key: IPrivateKey,
                 coin_conf: SubstrateCoinConf) -> None:
        """
        Construct class.

        Args:
            priv_key (IPrivateKey object) : Key object
            coin_conf (SubstrateCoinConf object): SubstrateCoinConf object

        Raises:
            SubstrateKeyError: If the bytes length is not valid
            TypeError: If the key is not a Sr25519PrivateKey object
        """
        if not isinstance(priv_key, Sr25519PrivateKey):
            raise TypeError("Invalid private key object type")
        self.m_priv_key = priv_key
        self.m_coin_conf = coin_conf

    def KeyObject(self) -> IPrivateKey:
        """
        Return the key object.

        Returns:
            IPrivateKey object: Key object
        """
        return self.m_priv_key

    @lru_cache()
    def Raw(self) -> DataBytes:
        """
        Return raw private key.

        Returns:
            DataBytes object: DataBytes object
        """
        return self.m_priv_key.Raw()

    @lru_cache()
    def PublicKey(self) -> SubstratePublicKey:
        """
        Get the public key correspondent to the private one.

        Returns:
            SubstratePublicKey object: SubstratePublicKey object
        """
        return SubstratePublicKey(self.m_priv_key.PublicKey(),
                                  self.m_coin_conf)

    @staticmethod
    def __KeyFromBytes(key_bytes: bytes) -> IPrivateKey:
        """
        Construct key from bytes.

        Args:
            key_bytes (bytes): Key bytes

        Returns:
            IPrivateKey object: IPrivateKey object

        Raises:
            SubstrateKeyError: If the key constructed from the bytes is not valid
        """
        try:
            return Sr25519PrivateKey.FromBytes(key_bytes)
        except ValueError as ex:
            raise SubstrateKeyError("Invalid private key") from ex



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for generic mnemonic validation."""

# Imports
from typing import Union

from bip_utils.utils.mnemonic.mnemonic import Mnemonic
from bip_utils.utils.mnemonic.mnemonic_decoder_base import MnemonicDecoderBase
from bip_utils.utils.mnemonic.mnemonic_ex import MnemonicChecksumError


class MnemonicValidator:
    """Mnemonic validator class."""

    m_mnemonic_decoder: MnemonicDecoderBase

    def __init__(self,
                 mnemonic_decoder: MnemonicDecoderBase) -> None:
        """
        Construct class.

        Args:
            mnemonic_decoder (MnemonicDecoderBase object): Mnemonic decoder class instance
        """
        self.m_mnemonic_decoder = mnemonic_decoder

    def Validate(self,
                 mnemonic: Union[str, Mnemonic]) -> None:
        """
        Validate the mnemonic specified at construction.

        Args:
            mnemonic (str or Mnemonic object): Mnemonic

        Raises:
            MnemonicChecksumError: If checksum is not valid
            ValueError: If mnemonic is not valid
        """

        # Just get entropy bytes without returning it, since it will validate the mnemonic
        self.m_mnemonic_decoder.Decode(mnemonic)

    def IsValid(self,
                mnemonic: Union[str, Mnemonic]) -> bool:
        """
        Get if the mnemonic specified at construction is valid.

        Args:
            mnemonic (str or Mnemonic object): Mnemonic

        Returns:
            bool: True if valid, False otherwise
        """

        # Simply try to validate
        try:
            self.Validate(mnemonic)
            return True
        except (ValueError, MnemonicChecksumError):
            return False



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for P2SH address encoding/decoding."""

# Imports
from typing import Any, Union

from bip_utils.addr.addr_key_validator import AddrKeyValidator
from bip_utils.addr.iaddr_decoder import IAddrDecoder
from bip_utils.addr.iaddr_encoder import IAddrEncoder
from bip_utils.addr.P2PKH_addr import BchP2PKHAddrDecoder, P2PKHAddrDecoder
from bip_utils.base58 import Base58Encoder
from bip_utils.bech32 import BchBech32Encoder
from bip_utils.ecc import IPublicKey
from bip_utils.utils.crypto import Hash160


class P2SHAddrConst:
    """Class container for P2SH constants."""

    # Script bytes
    SCRIPT_BYTES: bytes = b"\x00\x14"


class _P2SHAddrUtils:
    """Class container for P2SH utility functions."""

    @staticmethod
    def AddScriptSig(pub_key: IPublicKey) -> bytes:
        """
        Add script signature to public key and get address bytes.

        Args:
            pub_key (IPublicKey object): Public key object

        Returns:
            bytes: Address bytes
        """

        # Key hash: Hash160(public_key)
        key_hash_bytes = Hash160.QuickDigest(pub_key.RawCompressed().ToBytes())
        # Script signature: 0x0014 | Hash160(public_key)
        script_sig_bytes = P2SHAddrConst.SCRIPT_BYTES + key_hash_bytes
        # Address bytes = Hash160(script_signature)
        return Hash160.QuickDigest(script_sig_bytes)


class P2SHAddrDecoder(IAddrDecoder):
    """
    P2SH address decoder class.
    It allows the Pay-to-Script-Hash address decoding.
    """

    @staticmethod
    def DecodeAddr(addr: str,
                   **kwargs: Any) -> bytes:
        """
        Decode a P2SH address to bytes.

        Args:
            addr (str): Address string

        Other Parameters:
            net_ver (bytes): Expected net address version

        Returns:
            bytes: Script signature hash bytes

        Raises:
            ValueError: If the address encoding is not valid
        """

        # The decoding steps are the same of P2PKH
        return P2PKHAddrDecoder.DecodeAddr(addr, net_ver=kwargs["net_ver"])


class P2SHAddrEncoder(IAddrEncoder):
    """
    P2SH address encoder class.
    It allows the Pay-to-Script-Hash address encoding.
    """

    @staticmethod
    def EncodeKey(pub_key: Union[bytes, IPublicKey],
                  **kwargs: Any) -> str:
        """
        Encode a public key to P2SH address.

        Args:
            pub_key (bytes or IPublicKey) : Public key bytes or object

        Other Parameters:
            net_ver (bytes): Net address version

        Returns:
            str: Address string

        Raises:
            ValueError: If the public key is not valid
            TypeError: If the public key is not secp256k1
        """
        net_ver_bytes = kwargs["net_ver"]

        pub_key_obj = AddrKeyValidator.ValidateAndGetSecp256k1Key(pub_key)
        return Base58Encoder.CheckEncode(net_ver_bytes + _P2SHAddrUtils.AddScriptSig(pub_key_obj))


class BchP2SHAddrDecoder(IAddrDecoder):
    """
    Bitcoin Cash P2SH address decoder class.
    It allows the Bitcoin Cash P2SH decoding.
    """

    @staticmethod
    def DecodeAddr(addr: str,
                   **kwargs: Any) -> bytes:
        """
        Decode a Bitcoin Cash P2SH address to bytes.

        Args:
            addr (str): Address string

        Other Parameters:
            hrp (str)      : Expected HRP
            net_ver (bytes): Expected net address version

        Returns:
            bytes: Script signature hash bytes

        Raises:
            ValueError: If the address encoding is not valid
        """

        # The decoding steps are the same of P2PKH
        return BchP2PKHAddrDecoder.DecodeAddr(addr, hrp=kwargs["hrp"], net_ver=kwargs["net_ver"])


class BchP2SHAddrEncoder(IAddrEncoder):
    """
    Bitcoin Cash P2SH address encoder class.
    It allows the Bitcoin Cash P2SH encoding.
    """

    @staticmethod
    def EncodeKey(pub_key: Union[bytes, IPublicKey],
                  **kwargs: Any) -> str:
        """
        Encode a public key to Bitcoin Cash P2SH address.

        Args:
            pub_key (bytes or IPublicKey): Public key bytes or object

        Other Parameters:
            hrp (str)      : HRP
            net_ver (bytes): Net address version

        Returns:
            str: Address string

        Raises:
            ValueError: If the public key is not valid
            TypeError: If the public key is not secp256k1
        """
        hrp = kwargs["hrp"]
        net_ver_bytes = kwargs["net_ver"]

        pub_key_obj = AddrKeyValidator.ValidateAndGetSecp256k1Key(pub_key)
        return BchBech32Encoder.Encode(hrp, net_ver_bytes, _P2SHAddrUtils.AddScriptSig(pub_key_obj))


# Deprecated: only for compatibility, Encoder classes shall be used instead
P2SHAddr = P2SHAddrEncoder
BchP2SHAddr = BchP2SHAddrEncoder



from bip_utils.electrum.mnemonic_v2.electrum_v2_entropy_generator import (
    ElectrumV2EntropyBitLen, ElectrumV2EntropyGenerator
)
from bip_utils.electrum.mnemonic_v2.electrum_v2_mnemonic import (
    ElectrumV2Languages, ElectrumV2Mnemonic, ElectrumV2MnemonicTypes, ElectrumV2WordsNum
)
from bip_utils.electrum.mnemonic_v2.electrum_v2_mnemonic_decoder import ElectrumV2MnemonicDecoder
from bip_utils.electrum.mnemonic_v2.electrum_v2_mnemonic_encoder import ElectrumV2MnemonicEncoder
from bip_utils.electrum.mnemonic_v2.electrum_v2_mnemonic_generator import ElectrumV2MnemonicGenerator
from bip_utils.electrum.mnemonic_v2.electrum_v2_mnemonic_validator import ElectrumV2MnemonicValidator
from bip_utils.electrum.mnemonic_v2.electrum_v2_seed_generator import ElectrumV2SeedGenerator



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for Nano address encoding/decoding."""

# Imports
from typing import Any, Union

from bip_utils.addr.addr_dec_utils import AddrDecUtils
from bip_utils.addr.addr_key_validator import AddrKeyValidator
from bip_utils.addr.iaddr_decoder import IAddrDecoder
from bip_utils.addr.iaddr_encoder import IAddrEncoder
from bip_utils.coin_conf import CoinsConf
from bip_utils.ecc import Ed25519Blake2bPublicKey, IPublicKey
from bip_utils.utils.crypto import Blake2b40
from bip_utils.utils.misc import Base32Decoder, Base32Encoder, BytesUtils


class NanoAddrConst:
    """Class container for Nano address constants."""

    # Alphabet for base32
    BASE32_ALPHABET: str = "13456789abcdefghijkmnopqrstuwxyz"
    # Payload padding (decoded)
    PAYLOAD_PAD_DEC: bytes = b"\x00\x00\x00"
    # Payload padding (encoded)
    PAYLOAD_PAD_ENC: str = "1111"


class _NanoAddrUtils:
    """Nano address utility class."""

    @staticmethod
    def ComputeChecksum(pub_key_bytes: bytes) -> bytes:
        """
        Compute checksum in Nano format.

        Args:
            pub_key_bytes (bytes): Public key bytes

        Returns:
            bytes: Computed checksum
        """
        return BytesUtils.Reverse(Blake2b40.QuickDigest(pub_key_bytes))


class NanoAddrDecoder(IAddrDecoder):
    """
    Nano address decoder class.
    It allows the Nano address decoding.
    """

    @staticmethod
    def DecodeAddr(addr: str,
                   **kwargs: Any) -> bytes:
        """
        Decode a Nano address to bytes.

        Args:
            addr (str): Address string
            **kwargs  : Not used

        Returns:
            bytes: Public key bytes

        Raises:
            ValueError: If the address encoding is not valid
        """

        # Validate and remove prefix
        addr_no_prefix = AddrDecUtils.ValidateAndRemovePrefix(addr,
                                                              CoinsConf.Nano.ParamByKey("addr_prefix"))
        # Decode from base32
        addr_dec_bytes = Base32Decoder.Decode(NanoAddrConst.PAYLOAD_PAD_ENC + addr_no_prefix,
                                              NanoAddrConst.BASE32_ALPHABET)
        # Validate length
        AddrDecUtils.ValidateLength(addr_dec_bytes,
                                    Ed25519Blake2bPublicKey.CompressedLength() + Blake2b40.DigestSize()
                                    + len(NanoAddrConst.PAYLOAD_PAD_DEC) - 1)

        # Get back checksum and public key bytes
        pub_key_bytes, checksum_bytes = AddrDecUtils.SplitPartsByChecksum(
            addr_dec_bytes[len(NanoAddrConst.PAYLOAD_PAD_DEC):],
            Blake2b40.DigestSize()
        )
        # Validate checksum
        AddrDecUtils.ValidateChecksum(pub_key_bytes, checksum_bytes, _NanoAddrUtils.ComputeChecksum)
        # Validate public key
        AddrDecUtils.ValidatePubKey(pub_key_bytes, Ed25519Blake2bPublicKey)

        return pub_key_bytes


class NanoAddrEncoder(IAddrEncoder):
    """
    Nano address encoder class.
    It allows the Nano address encoding.
    """

    @staticmethod
    def EncodeKey(pub_key: Union[bytes, IPublicKey],
                  **kwargs: Any) -> str:
        """
        Encode a public key to Nano address.

        Args:
            pub_key (bytes or IPublicKey): Public key bytes or object
            **kwargs: Not used

        Returns:
            str: Address string

        Raises:
            ValueError: If the public key is not valid
            TypeError: If the public key is not ed25519-blake2b
        """
        pub_key_obj = AddrKeyValidator.ValidateAndGetEd25519Blake2bKey(pub_key)
        pub_key_bytes = pub_key_obj.RawCompressed().ToBytes()[1:]

        # Compute checksum
        checksum_bytes = _NanoAddrUtils.ComputeChecksum(pub_key_bytes)
        # Encode to base32
        payload_bytes = NanoAddrConst.PAYLOAD_PAD_DEC + pub_key_bytes + checksum_bytes
        b32_enc = Base32Encoder.EncodeNoPadding(payload_bytes, NanoAddrConst.BASE32_ALPHABET)

        # Add prefix
        return CoinsConf.Nano.ParamByKey("addr_prefix") + b32_enc[len(NanoAddrConst.PAYLOAD_PAD_ENC):]


# Deprecated: only for compatibility, Encoder class shall be used instead
NanoAddr = NanoAddrEncoder



from bip_utils.cardano.cip1852.conf.cip1852_coins import Cip1852Coins
from bip_utils.cardano.cip1852.conf.cip1852_conf import Cip1852Conf
from bip_utils.cardano.cip1852.conf.cip1852_conf_getter import Cip1852ConfGetter



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for Stellar address encoding/decoding."""

# Imports
from enum import IntEnum, unique
from typing import Any, Union

from bip_utils.addr.addr_dec_utils import AddrDecUtils
from bip_utils.addr.addr_key_validator import AddrKeyValidator
from bip_utils.addr.iaddr_decoder import IAddrDecoder
from bip_utils.addr.iaddr_encoder import IAddrEncoder
from bip_utils.ecc import Ed25519PublicKey, IPublicKey
from bip_utils.utils.crypto import XModemCrc
from bip_utils.utils.misc import Base32Decoder, Base32Encoder, BytesUtils, IntegerUtils


@unique
class XlmAddrTypes(IntEnum):
    """Enumerative for Stellar address types."""

    PUB_KEY = 6 << 3
    PRIV_KEY = 18 << 3


class XlmAddrConst:
    """Class container for Stellar address constants."""

    # Checksum length in bytes
    CHECKSUM_BYTE_LEN: int = 2


class _XlmAddrUtils:
    """Stellar address utility class."""

    @staticmethod
    def ComputeChecksum(payload_bytes: bytes) -> bytes:
        """
        Compute checksum in Stellar format.

        Args:
            payload_bytes (bytes): Payload bytes

        Returns:
            bytes: Computed checksum
        """
        return BytesUtils.Reverse(XModemCrc.QuickDigest(payload_bytes))


class XlmAddrDecoder(IAddrDecoder):
    """
    Stellar address decoder class.
    It allows the Stellar address decoding.
    """

    @staticmethod
    def DecodeAddr(addr: str,
                   **kwargs: Any) -> bytes:
        """
        Decode a Stellar address to bytes.

        Args:
            addr (str): Address string

        Other Parameters:
            addr_type (XlmAddrTypes): Expected address type (default: public key)

        Returns:
            bytes: Public key bytes

        Raises:
            ValueError: If the address encoding is not valid
            TypeError: If the address type is not a XlmAddrTypes enum
        """

        # Get and check address type
        addr_type = kwargs.get("addr_type", XlmAddrTypes.PUB_KEY)
        if not isinstance(addr_type, XlmAddrTypes):
            raise TypeError("Address type is not an enumerative of XlmAddrTypes")

        # Decode from base32
        addr_dec_bytes = Base32Decoder.Decode(addr)
        # Validate length
        AddrDecUtils.ValidateLength(addr_dec_bytes,
                                    Ed25519PublicKey.CompressedLength() + XlmAddrConst.CHECKSUM_BYTE_LEN)
        # Get back checksum and payload bytes
        payload_bytes, checksum_bytes = AddrDecUtils.SplitPartsByChecksum(addr_dec_bytes,
                                                                          XlmAddrConst.CHECKSUM_BYTE_LEN)
        # Check address type
        addr_type_got = payload_bytes[0]
        if addr_type != addr_type_got:
            raise ValueError(f"Invalid address type (expected {addr_type.value}, "
                             f"got {addr_type_got})")

        # Validate checksum
        AddrDecUtils.ValidateChecksum(payload_bytes, checksum_bytes, _XlmAddrUtils.ComputeChecksum)
        # Validate public key
        pub_key_bytes = payload_bytes[1:]
        AddrDecUtils.ValidatePubKey(pub_key_bytes, Ed25519PublicKey)

        return pub_key_bytes


class XlmAddrEncoder(IAddrEncoder):
    """
    Stellar address encoder class.
    It allows the Stellar address encoding.
    """

    @staticmethod
    def EncodeKey(pub_key: Union[bytes, IPublicKey],
                  **kwargs: Any) -> str:
        """
        Encode a public key to Stellar address.

        Args:
            pub_key (bytes or IPublicKey): Public key bytes or object

        Other Parameters:
            addr_type (XlmAddrTypes): Address type (default: public key)

        Returns:
            str: Address string

        Raises:
            ValueError: If the public key is not valid
            TypeError: If the public key is not ed25519 or address type is not a XlmAddrTypes enum
        """

        # Get and check address type
        addr_type = kwargs.get("addr_type", XlmAddrTypes.PUB_KEY)
        if not isinstance(addr_type, XlmAddrTypes):
            raise TypeError("Address type is not an enumerative of XlmAddrTypes")

        # Get public key
        pub_key_obj = AddrKeyValidator.ValidateAndGetEd25519Key(pub_key)
        payload_bytes = IntegerUtils.ToBytes(addr_type) + pub_key_obj.RawCompressed().ToBytes()[1:]

        # Compute checksum
        checksum_bytes = _XlmAddrUtils.ComputeChecksum(payload_bytes)
        # Encode to base32
        return Base32Encoder.EncodeNoPadding(payload_bytes + checksum_bytes)


# Deprecated: only for compatibility, Encoder class shall be used instead
XlmAddr = XlmAddrEncoder



# Copyright (c) 2022 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Helper library for ed25519 point encoding/decoding, which cannot be done with pynacl APIs.
Encode/Decode operations copied from: https://github.com/warner/python-pure25519/blob/master/pure25519/basic.py
"""
import binascii
from typing import Tuple, Union

from nacl import bindings

from bip_utils.utils.misc import BytesUtils, IntegerUtils


_Q = 2 ** 255 - 19
_L = 2**252 + 27742317777372353535851937790883648493
_G = (15112221349535400772501151409588531511454012693041857206046113283949847762202,
      46316835694926478169428394003475163141307993866256225615783033603165251855960)
_G_DEC_BYTES = binascii.unhexlify("1ad5258f602d56c9b2a7259560c72c695cdcd6fd31e2a4c0fe536ecdd3366921"
                                  "5866666666666666666666666666666666666666666666666666666666666666")
_G_ENC_BYTES = binascii.unhexlify("5866666666666666666666666666666666666666666666666666666666666666")
_COORD_BYTE_LEN = 32


def _inv(x: int) -> int:
    return pow(x, _Q - 2, _Q)


_D = -121665 * _inv(121666)
_I = pow(2, (_Q - 1) // 4, _Q)  # noqa: E741


def _x_recover(y: int) -> int:
    xx = (y * y - 1) * _inv(_D * y * y + 1)
    x = pow(xx, (_Q + 3) // 8, _Q)
    if (x * x - xx) % _Q != 0:
        x = (x * _I) % _Q
    if x % 2 != 0:
        x = _Q - x
    return x


def int_decode(int_bytes: bytes) -> int:
    """
    Decode int from bytes.

    Args:
        int_bytes (bytes): Integer bytes

    Returns:
        int: Decoded integer
    """
    return BytesUtils.ToInteger(int_bytes, endianness="little")


def int_encode(int_val: int) -> bytes:
    """
    Encode int to bytes.

    Args:
        int_val (int): Integer value

    Returns:
        bytes: Encoded integer
    """
    return IntegerUtils.ToBytes(int_val, _COORD_BYTE_LEN, endianness="little")


def point_is_decoded_bytes(point_bytes: bytes) -> bool:
    """
    Get if point bytes are in decoded format.

    Args:
        point_bytes (bytes): Point bytes

    Returns:
        bool: True if in decoded format, false otherwise
    """
    return len(point_bytes) == _COORD_BYTE_LEN * 2


def point_is_encoded_bytes(point_bytes: bytes) -> bool:
    """
    Get if point bytes are in encoded format.

    Args:
        point_bytes (bytes): Point bytes

    Returns:
        bool: True if in encoded format, false otherwise
    """
    return len(point_bytes) == _COORD_BYTE_LEN


def point_is_valid_bytes(point_bytes: bytes) -> bool:
    """
    Get if point bytes are valid.

    Args:
        point_bytes (bytes): Point bytes

    Returns:
        bool: True if valid, false otherwise
    """
    return point_is_decoded_bytes(point_bytes) or point_is_encoded_bytes(point_bytes)


def point_bytes_to_coord(point_bytes: bytes) -> Tuple[int, int]:
    """
    Convert point bytes to coordinates.

    Args:
        point_bytes (bytes): Point bytes

    Returns:
        tuple[int, int]: Point coordinates

    Raises:
        ValueError: If point bytes are not valid
    """
    if point_is_decoded_bytes(point_bytes):
        return int_decode(point_bytes[:_COORD_BYTE_LEN]), int_decode(point_bytes[_COORD_BYTE_LEN:])
    if point_is_encoded_bytes(point_bytes):
        return point_decode_no_check(point_bytes)
    raise ValueError("Invalid point bytes")


def point_coord_to_bytes(point_coord: Tuple[int, int]) -> bytes:
    """
    Convert point coordinates to bytes.

    Args:
        point_coord (tuple[int, int]): Point coordinates

    Returns:
        bytes: Point bytes
    """
    return int_encode(point_coord[0]) + int_encode(point_coord[1])


def point_decode_no_check(point_bytes: bytes) -> Tuple[int, int]:
    """
    Decode point bytes to coordinates without checking if it lies on the ed25519 curve.

    Args:
        point_bytes (bytes): Point bytes

    Returns:
        tuple[int, int]: Point coordinates

    Raises:
        ValueError: If point bytes are not valid
    """
    if not point_is_encoded_bytes(point_bytes):
        raise ValueError("Invalid point bytes")

    point_int = int_decode(point_bytes)

    clamp = (1 << 255) - 1
    y = point_int & clamp
    x = _x_recover(y)
    if bool(x & 1) != bool(point_int & (1 << 255)):
        x = _Q - x

    return x, y


def point_decode(point_bytes: bytes) -> Tuple[int, int]:
    """
    Decode point bytes to coordinates by checking if it lies on the ed25519 curve.

    Args:
        point_bytes (bytes): Point bytes

    Returns:
        tuple[int, int]: Point coordinates

    Raises:
        ValueError: If the point bytes are not valid or the decoded point doesn't lie on the curve
    """
    point_coord = point_decode_no_check(point_bytes)
    if not point_is_on_curve(point_coord):
        raise ValueError("Decoded point does not lie on the curve")
    return point_coord


def point_encode(point_coord: Tuple[int, int]) -> bytes:
    """
    Encode point coordinates to bytes.

    Args:
        point_coord (tuple[int, int]): Point coordinates

    Returns:
        bytes: Point bytes
    """
    point_bytes = point_coord_to_bytes(point_coord)

    y_bytes = bytearray(point_bytes[_COORD_BYTE_LEN:])
    if point_bytes[0] & 1:
        y_bytes[len(y_bytes) - 1] |= 0x80
    return bytes(y_bytes)


def point_is_generator(point: Union[bytes, Tuple[int, int]]) -> bool:
    """
    Get if the point is the generator of the ed25519 curve.

    Args:
        point (bytes or tuple[int, int]): Point

    Returns:
        bool: True if generator, false otherwise

    Raises:
        ValueError: If point bytes are not valid
    """
    # Avoid converting to coordinates if bytes to increase speed
    if isinstance(point, bytes):
        if point_is_encoded_bytes(point):
            return point == _G_ENC_BYTES
        if point_is_decoded_bytes(point):
            return point == _G_DEC_BYTES
        raise ValueError("Invalid point bytes")
    return point == _G


def point_is_on_curve(point: Union[bytes, Tuple[int, int]]) -> bool:
    """
    Get if the point lies on the ed25519 curve.
    This method is used because nacl.bindings.crypto_core_ed25519_is_valid_point performs more strict checks,
    which results in points (i.e. public keys) that are considered not valid even if they are accepted by wallets.

    Args:
        point (bytes or tuple[int, int]): Point

    Returns:
        bool: True if it lies on the curve, false otherwise

    Raises:
        ValueError: If point bytes are not valid
    """
    if isinstance(point, bytes):
        point = point_bytes_to_coord(point)

    x = point[0]
    y = point[1]
    return (-x * x + y * y - 1 - _D * x * x * y * y) % _Q == 0


def point_add(point_1: Union[bytes, Tuple[int, int]],
              point_2: Union[bytes, Tuple[int, int]]) -> bytes:
    """
    Add two points on the ed25519 curve.

    Args:
        point_1 (bytes or tuple[int, int]): Point 1
        point_2 (bytes or tuple[int, int]): Point 2

    Returns:
        bytes: New point resulting from the addition
    """
    return bindings.crypto_core_ed25519_add(
        point_1 if isinstance(point_1, bytes) else point_encode(point_1),
        point_2 if isinstance(point_2, bytes) else point_encode(point_2)
    )


def point_scalar_mul(scalar: Union[bytes, int],
                     point: Union[bytes, Tuple[int, int]]) -> bytes:
    """
    Multiply a point on the ed25519 curve with a scalar.

    Args:
        scalar (bytes or int)           : Scalar
        point (bytes or tuple[int, int]): Point

    Returns:
        bytes: New point resulting from the multiplication
    """
    return bindings.crypto_scalarmult_ed25519_noclamp(
        scalar if isinstance(scalar, bytes) else int_encode(scalar),
        point if isinstance(point, bytes) else point_encode(point)
    )


def point_scalar_mul_base(scalar: Union[bytes, int]) -> bytes:
    """
    Multiply the base (i.e. generator) point of the ed25519 curve with a scalar.

    Args:
        scalar (bytes or int): Scalar

    Returns:
        bytes: New point resulting from the multiplication
    """
    return bindings.crypto_scalarmult_ed25519_base_noclamp(
        scalar if isinstance(scalar, bytes) else int_encode(scalar)
    )


def scalar_reduce(scalar: Union[bytes, int]) -> bytes:
    """
    Convert the specified bytes to integer and return its lowest 32-bytes modulo ed25519 curve order.

    Args:
        scalar (bytes or int): Scalar

    Returns:
        bytes: Lowest 32-bytes modulo ed25519-order
    """
    if isinstance(scalar, int):
        scalar = int_encode(scalar)
    return bindings.crypto_core_ed25519_scalar_reduce(
        scalar.ljust(_COORD_BYTE_LEN * 2, b"\x00")
    )


def scalar_is_valid(scalar: Union[bytes, int]) -> bool:
    """
    Get if the specified scalar is valid (i.e. less than the ed25519 curve order).

    Args:
        scalar (bytes or int): Scalar

    Returns:
        bool: True if lower, false otherwise
    """
    if isinstance(scalar, bytes):
        scalar = int_decode(scalar)
    return scalar < _L



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for BIP39 mnemonic."""

# Imports
from enum import IntEnum, auto, unique
from typing import Dict, List, Union

from bip_utils.utils.misc import StringUtils
from bip_utils.utils.mnemonic import Mnemonic, MnemonicLanguages


@unique
class Bip39WordsNum(IntEnum):
    """Enumerative for BIP39 words number."""

    WORDS_NUM_12 = 12
    WORDS_NUM_15 = 15
    WORDS_NUM_18 = 18
    WORDS_NUM_21 = 21
    WORDS_NUM_24 = 24


@unique
class Bip39Languages(MnemonicLanguages):
    """Enumerative for BIP39 languages."""

    CHINESE_SIMPLIFIED = auto()
    CHINESE_TRADITIONAL = auto()
    CZECH = auto()
    ENGLISH = auto()
    FRENCH = auto()
    ITALIAN = auto()
    KOREAN = auto()
    PORTUGUESE = auto()
    SPANISH = auto()


class Bip39MnemonicConst:
    """Class container for BIP39 mnemonic constants."""

    # Accepted mnemonic word numbers
    MNEMONIC_WORD_NUM: List[Bip39WordsNum] = [
        Bip39WordsNum.WORDS_NUM_12,
        Bip39WordsNum.WORDS_NUM_15,
        Bip39WordsNum.WORDS_NUM_18,
        Bip39WordsNum.WORDS_NUM_21,
        Bip39WordsNum.WORDS_NUM_24,
    ]

    # Language files
    LANGUAGE_FILES: Dict[MnemonicLanguages, str] = {
        Bip39Languages.ENGLISH: "wordlist/english.txt",
        Bip39Languages.ITALIAN: "wordlist/italian.txt",
        Bip39Languages.FRENCH: "wordlist/french.txt",
        Bip39Languages.SPANISH: "wordlist/spanish.txt",
        Bip39Languages.PORTUGUESE: "wordlist/portuguese.txt",
        Bip39Languages.CZECH: "wordlist/czech.txt",
        Bip39Languages.CHINESE_SIMPLIFIED: "wordlist/chinese_simplified.txt",
        Bip39Languages.CHINESE_TRADITIONAL: "wordlist/chinese_traditional.txt",
        Bip39Languages.KOREAN: "wordlist/korean.txt",
    }

    # Total number of words
    WORDS_LIST_NUM: int = 2048
    # Word length in bit
    WORD_BIT_LEN: int = 11


class Bip39Mnemonic(Mnemonic):
    """
    BIP39 mnemonic class.
    It adds NFKD normalization to mnemonic.
    """

    @staticmethod
    def _Normalize(mnemonic: Union[str, List[str]]) -> List[str]:
        """
        Normalize mnemonic list.

        Args:
            mnemonic (str or list[str]): Mnemonic

        Returns:
            list[str]: Normalized mnemonic list
        """
        mnemonic = Mnemonic._Normalize(mnemonic)
        return list(map(lambda s: StringUtils.NormalizeNfkd(s.lower()), mnemonic))



from bip_utils.substrate.scale.substrate_scale_enc_base import SubstrateScaleEncoderBase
from bip_utils.substrate.scale.substrate_scale_enc_bytes import SubstrateScaleBytesEncoder
from bip_utils.substrate.scale.substrate_scale_enc_cuint import SubstrateScaleCUintEncoder
from bip_utils.substrate.scale.substrate_scale_enc_uint import (
    SubstrateScaleU8Encoder, SubstrateScaleU16Encoder, SubstrateScaleU32Encoder, SubstrateScaleU64Encoder,
    SubstrateScaleU128Encoder, SubstrateScaleU256Encoder
)



# Version
from bip_utils._version import __version__

# Address computation
from bip_utils.addr import (
    AdaByronAddrDecoder, AdaByronAddrTypes, AdaByronIcarusAddr, AdaByronIcarusAddrEncoder, AdaByronLegacyAddr,
    AdaByronLegacyAddrEncoder, AdaShelleyAddr, AdaShelleyAddrDecoder, AdaShelleyAddrEncoder, AdaShelleyAddrNetworkTags,
    AdaShelleyRewardAddr, AdaShelleyRewardAddrDecoder, AdaShelleyRewardAddrEncoder, AdaShelleyStakingAddr,
    AdaShelleyStakingAddrDecoder, AdaShelleyStakingAddrEncoder, AlgoAddr, AlgoAddrDecoder, AlgoAddrEncoder, AptosAddr,
    AptosAddrDecoder, AptosAddrEncoder, AtomAddr, AtomAddrDecoder, AtomAddrEncoder, AvaxPChainAddr,
    AvaxPChainAddrDecoder, AvaxPChainAddrEncoder, AvaxXChainAddr, AvaxXChainAddrDecoder, AvaxXChainAddrEncoder,
    BchAddrConverter, BchP2PKHAddr, BchP2PKHAddrDecoder, BchP2PKHAddrEncoder, BchP2SHAddr, BchP2SHAddrDecoder,
    BchP2SHAddrEncoder, EgldAddr, EgldAddrDecoder, EgldAddrEncoder, EosAddr, EosAddrDecoder, EosAddrEncoder,
    ErgoNetworkTypes, ErgoP2PKHAddr, ErgoP2PKHAddrDecoder, ErgoP2PKHAddrEncoder, EthAddr, EthAddrDecoder,
    EthAddrEncoder, FilSecp256k1Addr, FilSecp256k1AddrDecoder, FilSecp256k1AddrEncoder, IcxAddr, IcxAddrDecoder,
    IcxAddrEncoder, NanoAddr, NanoAddrDecoder, NanoAddrEncoder, NearAddr, NearAddrDecoder, NearAddrEncoder, NeoAddr,
    NeoAddrDecoder, NeoAddrEncoder, OkexAddr, OkexAddrDecoder, OkexAddrEncoder, OneAddr, OneAddrDecoder, OneAddrEncoder,
    P2PKHAddr, P2PKHAddrDecoder, P2PKHAddrEncoder, P2PKHPubKeyModes, P2SHAddr, P2SHAddrDecoder, P2SHAddrEncoder,
    P2TRAddr, P2TRAddrDecoder, P2TRAddrEncoder, P2WPKHAddr, P2WPKHAddrDecoder, P2WPKHAddrEncoder, SolAddr,
    SolAddrDecoder, SolAddrEncoder, SubstrateEd25519Addr, SubstrateEd25519AddrDecoder, SubstrateEd25519AddrEncoder,
    SubstrateSr25519Addr, SubstrateSr25519AddrDecoder, SubstrateSr25519AddrEncoder, TrxAddr, TrxAddrDecoder,
    TrxAddrEncoder, XlmAddr, XlmAddrDecoder, XlmAddrEncoder, XlmAddrTypes, XmrAddr, XmrAddrDecoder, XmrAddrEncoder,
    XmrIntegratedAddr, XmrIntegratedAddrDecoder, XmrIntegratedAddrEncoder, XrpAddr, XrpAddrDecoder, XrpAddrEncoder,
    XtzAddr, XtzAddrDecoder, XtzAddrEncoder, XtzAddrPrefixes, ZilAddr, ZilAddrDecoder, ZilAddrEncoder
)

# Algorand mnemonic
from bip_utils.algorand.mnemonic import (
    AlgorandEntropyBitLen, AlgorandEntropyGenerator, AlgorandLanguages, AlgorandMnemonic, AlgorandMnemonicDecoder,
    AlgorandMnemonicEncoder, AlgorandMnemonicGenerator, AlgorandMnemonicValidator, AlgorandSeedGenerator,
    AlgorandWordsNum
)

# Base58
from bip_utils.base58 import (
    Base58Alphabets, Base58ChecksumError, Base58Decoder, Base58Encoder, Base58XmrDecoder, Base58XmrEncoder
)

# Bech32
from bip_utils.bech32 import (
    BchBech32Decoder, BchBech32Encoder, Bech32ChecksumError, Bech32Decoder, Bech32Encoder, SegwitBech32Decoder,
    SegwitBech32Encoder
)

# BIP32
from bip_utils.bip.bip32 import (
    Bip32ChainCode, Bip32Depth, Bip32DeserializedKey, Bip32Ed25519Blake2bSlip, Bip32Ed25519Kholaw, Bip32Ed25519Slip,
    Bip32FingerPrint, Bip32KeyData, Bip32KeyDeserializer, Bip32KeyError, Bip32KeyIndex, Bip32KeyNetVersions,
    Bip32KholawEd25519, Bip32Nist256p1, Bip32Path, Bip32PathError, Bip32PathParser, Bip32PrivateKey,
    Bip32PrivateKeySerializer, Bip32PublicKey, Bip32PublicKeySerializer, Bip32Secp256k1, Bip32Slip10Ed25519,
    Bip32Slip10Ed25519Blake2b, Bip32Slip10Nist256p1, Bip32Slip10Secp256k1, Bip32Utils
)

# BIP38
from bip_utils.bip.bip38 import Bip38Decrypter, Bip38EcKeysGenerator, Bip38Encrypter, Bip38PubKeyModes

# BIP39
from bip_utils.bip.bip39 import (
    Bip39EntropyBitLen, Bip39EntropyGenerator, Bip39Languages, Bip39Mnemonic, Bip39MnemonicDecoder,
    Bip39MnemonicEncoder, Bip39MnemonicGenerator, Bip39MnemonicValidator, Bip39SeedGenerator, Bip39WordsNum
)
from bip_utils.bip.bip44 import Bip44

# BIP44/49/84
from bip_utils.bip.bip44_base import Bip44Changes, Bip44DepthError, Bip44Levels, Bip44PrivateKey, Bip44PublicKey
from bip_utils.bip.bip49 import Bip49
from bip_utils.bip.bip84 import Bip84
from bip_utils.bip.bip86 import Bip86

# BIP coins configuration
from bip_utils.bip.conf.bip44 import Bip44Coins, Bip44Conf, Bip44ConfGetter
from bip_utils.bip.conf.bip49 import Bip49Coins, Bip49Conf, Bip49ConfGetter
from bip_utils.bip.conf.bip84 import Bip84Coins, Bip84Conf, Bip84ConfGetter
from bip_utils.bip.conf.bip86 import Bip86Coins, Bip86Conf, Bip86ConfGetter

# Cardano
from bip_utils.cardano.bip32 import CardanoByronLegacyBip32, CardanoIcarusBip32
from bip_utils.cardano.byron import CardanoByronLegacy
from bip_utils.cardano.cip1852 import Cip1852
from bip_utils.cardano.cip1852.conf import Cip1852Coins, Cip1852Conf, Cip1852ConfGetter
from bip_utils.cardano.mnemonic import CardanoByronLegacySeedGenerator, CardanoIcarusSeedGenerator
from bip_utils.cardano.shelley import CardanoShelley, CardanoShelleyPrivateKeys, CardanoShelleyPublicKeys

# Generic coins configuration
from bip_utils.coin_conf import CoinsConf

# ECC
from bip_utils.ecc import (
    Ed25519, Ed25519Blake2b, Ed25519Blake2bPoint, Ed25519Blake2bPrivateKey, Ed25519Blake2bPublicKey, Ed25519Kholaw,
    Ed25519KholawPoint, Ed25519KholawPrivateKey, Ed25519KholawPublicKey, Ed25519Monero, Ed25519MoneroPoint,
    Ed25519MoneroPrivateKey, Ed25519MoneroPublicKey, Ed25519Point, Ed25519PrivateKey, Ed25519PublicKey,
    EllipticCurveGetter, EllipticCurveTypes, IPoint, IPrivateKey, IPublicKey, Nist256p1, Nist256p1Point,
    Nist256p1PrivateKey, Nist256p1PublicKey, Secp256k1, Secp256k1Point, Secp256k1PrivateKey, Secp256k1PublicKey,
    Sr25519, Sr25519Point, Sr25519PrivateKey, Sr25519PublicKey
)

# Electrum wallet
from bip_utils.electrum import ElectrumV1, ElectrumV2Segwit, ElectrumV2Standard

# Electrum mnemonic
from bip_utils.electrum.mnemonic_v1 import (
    ElectrumV1EntropyBitLen, ElectrumV1EntropyGenerator, ElectrumV1Languages, ElectrumV1Mnemonic,
    ElectrumV1MnemonicDecoder, ElectrumV1MnemonicEncoder, ElectrumV1MnemonicGenerator, ElectrumV1MnemonicValidator,
    ElectrumV1SeedGenerator, ElectrumV1WordsNum
)
from bip_utils.electrum.mnemonic_v2 import (
    ElectrumV2EntropyBitLen, ElectrumV2EntropyGenerator, ElectrumV2Languages, ElectrumV2Mnemonic,
    ElectrumV2MnemonicDecoder, ElectrumV2MnemonicEncoder, ElectrumV2MnemonicGenerator, ElectrumV2MnemonicTypes,
    ElectrumV2MnemonicValidator, ElectrumV2SeedGenerator, ElectrumV2WordsNum
)

# Monero
from bip_utils.monero import Monero, MoneroKeyError, MoneroPrivateKey, MoneroPublicKey, MoneroSubaddress

# Monero configuration
from bip_utils.monero.conf import MoneroCoins, MoneroConf

# Monero mnemonic
from bip_utils.monero.mnemonic import (
    MoneroEntropyBitLen, MoneroEntropyGenerator, MoneroLanguages, MoneroMnemonic, MoneroMnemonicDecoder,
    MoneroMnemonicEncoder, MoneroMnemonicGenerator, MoneroMnemonicNoChecksumEncoder, MoneroMnemonicValidator,
    MoneroMnemonicWithChecksumEncoder, MoneroSeedGenerator, MoneroWordsNum
)

# SLIP32
from bip_utils.slip.slip32 import (
    Slip32DeserializedKey, Slip32KeyDeserializer, Slip32PrivateKeySerializer, Slip32PublicKeySerializer
)

# Solana
from bip_utils.solana import SplToken

# SS58
from bip_utils.ss58 import SS58ChecksumError, SS58Decoder, SS58Encoder

# Substrate
from bip_utils.substrate import (
    Substrate, SubstrateKeyError, SubstratePath, SubstratePathElem, SubstratePathError, SubstratePathParser,
    SubstratePrivateKey, SubstratePublicKey
)

# Substrate configuration
from bip_utils.substrate.conf import SubstrateCoins, SubstrateConf

# Substrate mnemonic
from bip_utils.substrate.mnemonic import SubstrateBip39SeedGenerator

# Substrate SCALE
from bip_utils.substrate.scale import (
    SubstrateScaleBytesEncoder, SubstrateScaleCUintEncoder, SubstrateScaleU8Encoder, SubstrateScaleU16Encoder,
    SubstrateScaleU32Encoder, SubstrateScaleU64Encoder, SubstrateScaleU128Encoder, SubstrateScaleU256Encoder
)

# Utils
from bip_utils.utils.crypto import (
    AesEcbDecrypter, AesEcbEncrypter, Blake2b, Blake2b160, Blake2b224, Blake2b256, ChaCha20Poly1305, Crc32,
    DoubleSha256, Hash160, HmacSha256, HmacSha512, Kekkak256, Pbkdf2HmacSha512, Ripemd160, Scrypt, Sha3_256, Sha256,
    Sha512, Sha512_256, XModemCrc
)
from bip_utils.utils.misc import AlgoUtils, BitUtils, BytesUtils, DataBytes, IntegerUtils, StringUtils
from bip_utils.utils.mnemonic import MnemonicChecksumError

# WIF
from bip_utils.wif import WifDecoder, WifEncoder, WifPubKeyModes



# Copyright (c) 2022 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Module for Cardano Byron legacy BIP32 keys derivation.

References:
    https://input-output-hk.github.io/cardano-wallet/concepts/master-key-generation
    https://cips.cardano.org/cips/cip3/byron.md
"""

# Imports
from bip_utils.bip.bip32 import Bip32KeyIndex, Bip32KholawEd25519KeyDerivatorBase, Bip32PublicKey
from bip_utils.ecc import EllipticCurve, IPoint
from bip_utils.utils.misc import BytesUtils, IntegerUtils


class CardanoByronLegacyKeyDerivator(Bip32KholawEd25519KeyDerivatorBase):
    """
    Cardano Byron legacy key derivator class.
    It allows keys derivation for Cardano-Byron (legacy, used by old versions of Daedalus).
    Derivation based on BIP32 ed25519 Khovratovich/Law with some differences on keys computation.
    """

    @staticmethod
    def _SerializeIndex(index: Bip32KeyIndex) -> bytes:
        """
        Serialize key index.

        Args:
            index (Bip32KeyIndex object): Key index

        Returns:
            bytes: Serialized index
        """
        return index.ToBytes(endianness="big")

    @staticmethod
    def _NewPrivateKeyLeftPart(zl_bytes: bytes,
                               kl_bytes: bytes,
                               curve: EllipticCurve) -> bytes:
        """
        Compute the new private key left part for private derivation.

        Args:
            zl_bytes (bytes)            : Leftmost Z 32-byte
            kl_bytes (bytes)            : Leftmost private key 32-byte
            curve (EllipticCurve object): EllipticCurve object

        Returns:
            bytes: Leftmost new private key 32-byte
        """
        zl8_bytes = BytesUtils.MultiplyScalarNoCarry(zl_bytes, 8)

        zl8_int = BytesUtils.ToInteger(zl8_bytes, endianness="little")
        kl_int = BytesUtils.ToInteger(kl_bytes, endianness="little")

        return IntegerUtils.ToBytes((zl8_int + kl_int) % curve.Order(), bytes_num=32, endianness="little")

    @staticmethod
    def _NewPrivateKeyRightPart(zr_bytes: bytes,
                                kr_bytes: bytes) -> bytes:
        """
        Compute the new private key right part for private derivation.

        Args:
            zr_bytes (bytes): Rightmost Z 32-byte
            kr_bytes (bytes): Rightmost private key 32-byte

        Returns:
            bytes: Rightmost new private key 32-byte
        """
        return BytesUtils.AddNoCarry(zr_bytes, kr_bytes)

    @staticmethod
    def _NewPublicKeyPoint(pub_key: Bip32PublicKey,
                           zl_bytes: bytes) -> IPoint:
        """
        Compute new public key point for public derivation.

        Args:
            pub_key (Bip32PublicKey object): Bip32PublicKey object
            zl_bytes (bytes)              : Leftmost Z 32-byte

        Returns:
            IPoint object: IPoint object
        """

        # Compute the new public key point: PKEY + 8ZL * G
        zl8_int = BytesUtils.ToInteger(BytesUtils.MultiplyScalarNoCarry(zl_bytes, 8),
                                       endianness="little")
        return pub_key.Point() + (zl8_int * pub_key.Curve().Generator())



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Module for P2TR address encoding/decoding.

References:
    https://github.com/bitcoin/bips/blob/master/bip-0340.mediawiki
    https://github.com/bitcoin/bips/blob/master/bip-0341.mediawiki
"""

# Imports
from typing import Any, Union

from bip_utils.addr.addr_dec_utils import AddrDecUtils
from bip_utils.addr.addr_key_validator import AddrKeyValidator
from bip_utils.addr.iaddr_decoder import IAddrDecoder
from bip_utils.addr.iaddr_encoder import IAddrEncoder
from bip_utils.bech32 import Bech32ChecksumError, SegwitBech32Decoder, SegwitBech32Encoder
from bip_utils.ecc import IPoint, IPublicKey, Secp256k1, Secp256k1Point, Secp256k1PublicKey
from bip_utils.utils.crypto import Sha256
from bip_utils.utils.misc import BytesUtils, IntegerUtils


class P2TRConst:
    """Class container for P2TR constants."""

    # Secp256k1 field size
    FIELD_SIZE: int = 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEFFFFFC2F
    # SHA256 of "TapTweak"
    TAP_TWEAK_SHA256: bytes = BytesUtils.FromHexString(
        "e80fe1639c9ca050e3af1b39c143c63e429cbceb15d940fbb5c5a1f4af57c5e9"
    )
    # Witness version is fixed to one for P2TR
    WITNESS_VER: int = 1


class _P2TRUtils:
    """Class container for P2TR utility functions."""

    @staticmethod
    def TaggedHash(tag: Union[bytes, str],
                   data_bytes: bytes) -> bytes:
        """
        Implementation of the hash tag function as defined by BIP-0340.
        Tagged hash = SHA256(SHA256(tag) || SHA256(tag) || data)

        Args:
            tag (bytes or str): Tag, if bytes it'll be considered already hashed
            data_bytes (bytes): Data bytes

        Returns:
            bytes: Tagged hash
        """
        tag_hash = Sha256.QuickDigest(tag) if isinstance(tag, str) else tag
        return Sha256.QuickDigest(tag_hash + tag_hash + data_bytes)

    @staticmethod
    def HashTapTweak(pub_key: IPublicKey) -> bytes:
        """
        Compute the HashTapTweak of the specified public key.

        Args:
            pub_key (IPublicKey object): Public key

        Returns:
            bytes: Computed hash
        """

        # Use the pre-computed SHA256 of "TapTweak" for speeding up
        return _P2TRUtils.TaggedHash(
            P2TRConst.TAP_TWEAK_SHA256,
            IntegerUtils.ToBytes(pub_key.Point().X(),
                                 bytes_num=Secp256k1Point.CoordinateLength())
        )

    @staticmethod
    def LiftX(pub_key: IPublicKey) -> IPoint:
        """
        Implementation of the lift_x function as defined by BIP-0340.
        It computes the point P for which P.X() = pub_key.X() and has_even_y(P).

        Args:
            pub_key (IPublicKey object): Public key

        Returns:
            IPoint: Computed point

        Raises:
            ValueError: If the point doesn't exist
        """
        p = P2TRConst.FIELD_SIZE
        x = pub_key.Point().X()
        if x >= p:
            raise ValueError("Unable to compute LiftX point")
        c = (pow(x, 3, p) + 7) % p
        y = pow(c, (p + 1) // 4, p)
        if c != pow(y, 2, p):
            raise ValueError("Unable to compute LiftX point")
        return Secp256k1Point.FromCoordinates(x, y if y % 2 == 0 else p - y)

    @staticmethod
    def TweakPublicKey(pub_key: IPublicKey) -> bytes:
        """
        Tweak a public key as defined by BIP-0086.
        tweaked_pub_key = lift_x(pub_key.X()) + int(HashTapTweak(bytes(pub_key.X()))) * G

        Args:
            pub_key (IPublicKey object): Public key

        Returns:
            bytes: X coordinate of the tweaked public key
        """
        h = _P2TRUtils.HashTapTweak(pub_key)
        out_point = _P2TRUtils.LiftX(pub_key) + (BytesUtils.ToInteger(h) * Secp256k1.Generator())
        return IntegerUtils.ToBytes(out_point.X(),
                                    bytes_num=Secp256k1Point.CoordinateLength())


class P2TRAddrDecoder(IAddrDecoder):
    """
    P2WPKH address decoder class.
    It allows the Pay-to-Witness-Public-Key-Hash address decoding.
    """

    @staticmethod
    def DecodeAddr(addr: str,
                   **kwargs: Any) -> bytes:
        """
        Decode a P2TR address to bytes.

        Args:
            addr (str): Address string

        Other Parameters:
            hrp (str): Expected HRP

        Returns:
            bytes: X coordinate of the tweaked public key

        Raises:
            ValueError: If the address encoding is not valid
        """
        hrp = kwargs["hrp"]

        try:
            wit_ver_got, addr_dec_bytes = SegwitBech32Decoder.Decode(hrp, addr)
        except Bech32ChecksumError as ex:
            raise ValueError("Invalid bech32 checksum") from ex

        # Validate length
        AddrDecUtils.ValidateLength(addr_dec_bytes, Secp256k1PublicKey.CompressedLength() - 1)
        # Check witness version
        if wit_ver_got != P2TRConst.WITNESS_VER:
            raise ValueError(f"Invalid witness version (expected {P2TRConst.WITNESS_VER}, got {wit_ver_got})")
        return addr_dec_bytes


class P2TRAddrEncoder(IAddrEncoder):
    """
    P2TR address encoder class.
    It allows the Pay-to-Taproot address encoding.
    """

    @staticmethod
    def EncodeKey(pub_key: Union[bytes, IPublicKey],
                  **kwargs: Any) -> str:
        """
        Encode a public key to P2TR address.

        Args:
            pub_key (bytes or IPublicKey): Public key bytes or object

        Other Parameters:
            hrp (str): HRP

        Returns:
            str: Address string

        Raises:
            ValueError: If the public key is not valid or cannot be tweaked
            TypeError: If the public key is not secp256k1
        """
        hrp = kwargs["hrp"]

        pub_key_obj = AddrKeyValidator.ValidateAndGetSecp256k1Key(pub_key)
        return SegwitBech32Encoder.Encode(hrp,
                                          P2TRConst.WITNESS_VER,
                                          _P2TRUtils.TweakPublicKey(pub_key_obj))


# Deprecated: only for compatibility, Encoder class shall be used instead
P2TRAddr = P2TRAddrEncoder



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Module for Electrum v2 mnemonic encoding.
Reference: https://github.com/spesmilo/electrum
"""

from bip_utils.bip.bip39.bip39_mnemonic_utils import Bip39WordsListGetter

# Imports
from bip_utils.electrum.mnemonic_v2.electrum_v2_entropy_generator import ElectrumV2EntropyGenerator
from bip_utils.electrum.mnemonic_v2.electrum_v2_mnemonic import (
    ElectrumV2Languages, ElectrumV2Mnemonic, ElectrumV2MnemonicTypes
)
from bip_utils.electrum.mnemonic_v2.electrum_v2_mnemonic_utils import ElectrumV2MnemonicUtils
from bip_utils.utils.misc import BytesUtils
from bip_utils.utils.mnemonic import Mnemonic, MnemonicEncoderBase


class ElectrumV2MnemonicEncoder(MnemonicEncoderBase):
    """
    Electrum v2 mnemonic encoder class.
    It encodes bytes to the mnemonic phrase.
    """

    m_mnemonic_type: ElectrumV2MnemonicTypes

    def __init__(self,
                 mnemonic_type: ElectrumV2MnemonicTypes,
                 lang: ElectrumV2Languages = ElectrumV2Languages.ENGLISH) -> None:
        """
        Construct class.

        Args:
            mnemonic_type (ElectrumV2MnemonicTypes): Mnemonic type
            lang (ElectrumV2Languages, optional)   : Language (default: English)

        Raises:
            TypeError: If the language is not a ElectrumV2Languages enum or
                       the mnemonic type is not a ElectrumV2MnemonicTypes enum
            ValueError: If loaded words list is not valid
        """
        if not isinstance(mnemonic_type, ElectrumV2MnemonicTypes):
            raise TypeError("Mnemonic type is not an enumerative of ElectrumV2MnemonicTypes")
        if not isinstance(lang, ElectrumV2Languages):
            raise TypeError("Language is not an enumerative of ElectrumV2Languages")
        super().__init__(lang.value, Bip39WordsListGetter)
        self.m_mnemonic_type = mnemonic_type

    def Encode(self,
               entropy_bytes: bytes) -> Mnemonic:
        """
        Encode bytes to mnemonic phrase.

        Args:
            entropy_bytes (bytes): Entropy bytes

        Returns:
            Mnemonic object: Encoded mnemonic

        Raises:
            ValueError: If bytes length is not valid or a mnemonic cannot be generated
        """

        # Check entropy length
        entropy_int = BytesUtils.ToInteger(entropy_bytes)
        if not ElectrumV2EntropyGenerator.AreEntropyBitsEnough(entropy_int):
            raise ValueError("Entropy bit length is not enough for generating a valid mnemonic")

        # Encode to words
        n = self.m_words_list.Length()
        mnemonic = []
        while entropy_int > 0:
            word_idx = entropy_int % n
            entropy_int //= n
            mnemonic.append(self.m_words_list.GetWordAtIdx(word_idx))

        # Check if the mnemonic is valid
        mnemonic_obj = ElectrumV2Mnemonic.FromList(mnemonic)
        if not ElectrumV2MnemonicUtils.IsValidMnemonic(mnemonic_obj, self.m_mnemonic_type):
            raise ValueError("Entropy bytes are not suitable for generating a valid mnemonic")

        return mnemonic_obj



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Module for Substrate coins configuration.
Reference: https://wiki.polkadot.network/docs/build-ss58-registry
"""

# Imports
from bip_utils.coin_conf import CoinsConf
from bip_utils.substrate.conf.substrate_coin_conf import SubstrateCoinConf


class SubstrateConf:
    """Class container for Substrate configuration."""

    # Configuration for Acala
    Acala: SubstrateCoinConf = SubstrateCoinConf.FromCoinConf(CoinsConf.Acala)

    # Configuration for Bifrost
    Bifrost: SubstrateCoinConf = SubstrateCoinConf.FromCoinConf(CoinsConf.Bifrost)

    # Configuration for ChainX
    ChainX: SubstrateCoinConf = SubstrateCoinConf.FromCoinConf(CoinsConf.ChainX)

    # Configuration for Edgeware
    Edgeware: SubstrateCoinConf = SubstrateCoinConf.FromCoinConf(CoinsConf.Edgeware)

    # Configuration for generic Substrate coin
    Generic: SubstrateCoinConf = SubstrateCoinConf.FromCoinConf(CoinsConf.GenericSubstrate)

    # Configuration for Karura
    Karura: SubstrateCoinConf = SubstrateCoinConf.FromCoinConf(CoinsConf.Karura)

    # Configuration for Kusama
    Kusama: SubstrateCoinConf = SubstrateCoinConf.FromCoinConf(CoinsConf.Kusama)

    # Configuration for Moonbeam
    Moonbeam: SubstrateCoinConf = SubstrateCoinConf.FromCoinConf(CoinsConf.Moonbeam)

    # Configuration for Moonriver
    Moonriver: SubstrateCoinConf = SubstrateCoinConf.FromCoinConf(CoinsConf.Moonriver)

    # Configuration for Phala
    Phala: SubstrateCoinConf = SubstrateCoinConf.FromCoinConf(CoinsConf.Phala)

    # Configuration for Plasm
    Plasm: SubstrateCoinConf = SubstrateCoinConf.FromCoinConf(CoinsConf.Plasm)

    # Configuration for Polkadot
    Polkadot: SubstrateCoinConf = SubstrateCoinConf.FromCoinConf(CoinsConf.Polkadot)

    # Configuration for Sora
    Sora: SubstrateCoinConf = SubstrateCoinConf.FromCoinConf(CoinsConf.Sora)

    # Configuration for Stafi
    Stafi: SubstrateCoinConf = SubstrateCoinConf.FromCoinConf(CoinsConf.Stafi)



from bip_utils.bech32.bch_bech32 import BchBech32Decoder, BchBech32Encoder
from bip_utils.bech32.bech32 import Bech32Decoder, Bech32Encoder
from bip_utils.bech32.bech32_ex import Bech32ChecksumError
from bip_utils.bech32.segwit_bech32 import SegwitBech32Decoder, SegwitBech32Encoder



# Copyright (c) 2022 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Module for Electrum v1 mnemonic encoding.
Reference: https://github.com/spesmilo/electrum
"""

# Imports
from bip_utils.electrum.mnemonic_v1.electrum_v1_entropy_generator import ElectrumV1EntropyGenerator
from bip_utils.electrum.mnemonic_v1.electrum_v1_mnemonic import ElectrumV1Languages, ElectrumV1Mnemonic
from bip_utils.electrum.mnemonic_v1.electrum_v1_mnemonic_utils import ElectrumV1WordsListGetter
from bip_utils.utils.mnemonic import Mnemonic, MnemonicEncoderBase, MnemonicUtils


class ElectrumV1MnemonicEncoder(MnemonicEncoderBase):
    """
    Electrum v1 mnemonic encoder class.
    It encodes bytes to the mnemonic phrase.
    """

    def __init__(self,
                 lang: ElectrumV1Languages = ElectrumV1Languages.ENGLISH) -> None:
        """
        Construct class.

        Args:
            lang (ElectrumV1Languages, optional): Language (default: English)

        Raises:
            TypeError: If the language is not a ElectrumV1Languages enum
            ValueError: If loaded words list is not valid
        """
        super().__init__(lang, ElectrumV1WordsListGetter)

    def Encode(self,
               entropy_bytes: bytes) -> Mnemonic:
        """
        Encode bytes to mnemonic phrase.

        Args:
            entropy_bytes (bytes): Entropy bytes (accepted lengths in bits: 128)

        Returns:
            Mnemonic object: Encoded mnemonic

        Raises:
            ValueError: If bytes length is not valid
        """

        # Check entropy length
        entropy_byte_len = len(entropy_bytes)
        if not ElectrumV1EntropyGenerator.IsValidEntropyByteLen(entropy_byte_len):
            raise ValueError(f"Entropy byte length ({entropy_byte_len}) is not valid")

        # Build mnemonic
        mnemonic = []
        for i in range(len(entropy_bytes) // 4):
            mnemonic += MnemonicUtils.BytesChunkToWords(entropy_bytes[i * 4:(i * 4) + 4], self.m_words_list, "big")

        return ElectrumV1Mnemonic.FromList(mnemonic)



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for getting account addresses of SPL tokens."""

#
# Imports
#
from typing import List

from bip_utils.addr import SolAddrDecoder
from bip_utils.base58 import Base58Encoder
from bip_utils.ecc import Ed25519PublicKey
from bip_utils.utils.crypto import Sha256
from bip_utils.utils.misc import IntegerUtils


#
# Classes
#

class SplTokenConst:
    """Class container for SPL token constants."""

    # Default program ID
    DEF_PROGRAM_ID: str = "ATokenGPvbdGVxr1b2hvZbsiqW5xWH25efTNsLJA8knL"
    # Default token program ID
    DEF_TOKEN_PROGRAM_ID: str = "TokenkegQfeZyiNwAJbNbGKPFXCWuBvf9Ss623VQ5DA"
    # Program derived address marker
    PDA_MARKER: bytes = b"ProgramDerivedAddress"
    # Maximum seed bump value
    SEED_BUMP_MAX_VAL: int = 2 ** 8 - 1
    # Maximum number of seeds
    SEEDS_MAX_NUM: int = 16


class SplToken:
    """
    SPL token class.
    It provides methods for getting the account address associated to a SPL token.
    """

    @classmethod
    def GetAssociatedTokenAddress(cls,
                                  wallet_addr: str,
                                  token_mint_addr: str) -> str:
        """
        Get the account address associated to the specified SPL token.

        Args:
            wallet_addr (str)    : Wallet address
            token_mint_addr (str): Token mint address

        Returns:
            str: Associated account address

        Raises:
            ValueError: If the account address cannot be found or the specified addresses are not valid
        """
        return cls.GetAssociatedTokenAddressWithProgramId(
            wallet_addr,
            token_mint_addr,
            SplTokenConst.DEF_TOKEN_PROGRAM_ID
        )

    @classmethod
    def GetAssociatedTokenAddressWithProgramId(cls,
                                               wallet_addr: str,
                                               token_mint_addr: str,
                                               token_program_id: str) -> str:
        """
        Get the account address associated to the specified SPL token and token program ID.

        Args:
            wallet_addr (str)     : Wallet address
            token_mint_addr (str) : Token mint address
            token_program_id (str): Token program ID

        Returns:
            str: Associated account address

        Raises:
            ValueError: If the account address cannot be found or the specified addresses or ID are not valid
        """
        seeds = [
            SolAddrDecoder.DecodeAddr(wallet_addr),
            SolAddrDecoder.DecodeAddr(token_program_id),
            SolAddrDecoder.DecodeAddr(token_mint_addr),
        ]
        return cls.FindPda(seeds, SplTokenConst.DEF_PROGRAM_ID)

    @classmethod
    def FindPda(cls,
                seeds: List[bytes],
                program_id: str) -> str:
        """
        Find a valid PDA (Program Derived Address) and its corresponding bump seed.

        Args:
            seeds (list[bytes]): List of seeds bytes
            program_id (str)   : Program ID

        Returns:
            str: Found PDA

        Raises:
            ValueError: If the PDA cannot be found or the specified seeds or program ID are not valid
        """

        # Check if seeds are valid
        if len(seeds) > SplTokenConst.SEEDS_MAX_NUM:
            raise ValueError(f"Seeds length is not valid ({len(seeds)})")
        for seed in seeds:
            if len(seed) > Ed25519PublicKey.CompressedLength() - 1:
                raise ValueError(f"Seed length is not valid ({len(seeds)})")

        program_id_bytes = SolAddrDecoder.DecodeAddr(program_id)
        bump_seed = SplTokenConst.SEED_BUMP_MAX_VAL
        for _ in range(SplTokenConst.SEED_BUMP_MAX_VAL):
            # Add bump to seeds
            seeds_with_bump = list(seeds)
            seeds_with_bump.append(IntegerUtils.ToBytes(bump_seed))
            # Try to create PDA
            try:
                return cls.__CreatePda(seeds_with_bump, program_id_bytes)
            except ValueError:
                # Continue with the next bump seed if PDA is not valid
                bump_seed -= 1

        # Very unlucky case
        raise ValueError("Unable to find a valid PDA")

    @staticmethod
    def __CreatePda(seeds_with_bump: List[bytes],
                    program_id_bytes: bytes) -> str:
        """
        Create a PDA (Program Derived Address) for the specified seeds and program ID.

        Args:
            seeds_with_bump (list[bytes]): List of seeds bytes with bump
            program_id_bytes (bytes)     : Program ID bytes

        Returns:
            str: Created PDA

        Raises:
            ValueError: If the created PDA is not valid
        """
        sha256 = Sha256()
        # Compute SHA256 of seeds with bump
        for seed in seeds_with_bump:
            sha256.Update(seed)
        # Compute SHA256 of program ID and PDA marker
        for elem in (program_id_bytes, SplTokenConst.PDA_MARKER):
            sha256.Update(elem)
        # Get PDA bytes
        pda_bytes = sha256.Digest()

        # A PDA shall NOT lie on the ed25519 curve, so it shall not be a valid public key
        if Ed25519PublicKey.IsValidBytes(pda_bytes):
            raise ValueError("Invalid created PDA")

        return Base58Encoder.Encode(pda_bytes)



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Module for BIP39 mnemonic encoding.
Reference: https://github.com/bitcoin/bips/blob/master/bip-0039.mediawiki
"""

# Imports
from bip_utils.bip.bip39.bip39_entropy_generator import Bip39EntropyGenerator
from bip_utils.bip.bip39.bip39_mnemonic import Bip39Languages, Bip39Mnemonic, Bip39MnemonicConst
from bip_utils.bip.bip39.bip39_mnemonic_utils import Bip39WordsListGetter
from bip_utils.utils.crypto import Sha256
from bip_utils.utils.misc import BytesUtils, IntegerUtils
from bip_utils.utils.mnemonic import Mnemonic, MnemonicEncoderBase


class Bip39MnemonicEncoder(MnemonicEncoderBase):
    """
    BIP39 mnemonic encoder class.
    It encodes bytes to the mnemonic phrase.
    """

    def __init__(self,
                 lang: Bip39Languages = Bip39Languages.ENGLISH) -> None:
        """
        Construct class.

        Args:
            lang (Bip39Languages, optional): Language (default: English)

        Raises:
            TypeError: If the language is not a Bip39Languages enum
            ValueError: If loaded words list is not valid
        """
        super().__init__(lang, Bip39WordsListGetter)

    def Encode(self,
               entropy_bytes: bytes) -> Mnemonic:
        """
        Encode bytes to mnemonic phrase.

        Args:
            entropy_bytes (bytes): Entropy bytes (accepted lengths in bits: 128, 160, 192, 224, 256)

        Returns:
            Mnemonic object: Encoded mnemonic

        Raises:
            ValueError: If entropy is not valid
        """

        # Check entropy length
        entropy_byte_len = len(entropy_bytes)
        if not Bip39EntropyGenerator.IsValidEntropyByteLen(entropy_byte_len):
            raise ValueError(f"Entropy byte length ({entropy_byte_len}) is not valid")

        # Convert entropy to binary string
        entropy_bin_str = BytesUtils.ToBinaryStr(entropy_bytes, entropy_byte_len * 8)
        # Get entropy hash as binary string
        entropy_hash_bin_str = BytesUtils.ToBinaryStr(Sha256.QuickDigest(entropy_bytes),
                                                      Sha256.DigestSize() * 8)
        # Get mnemonic binary string by concatenating entropy and checksum
        mnemonic_bin_str = entropy_bin_str + entropy_hash_bin_str[:entropy_byte_len // 4]

        # Get mnemonic from entropy
        mnemonic = []
        for i in range(len(mnemonic_bin_str) // Bip39MnemonicConst.WORD_BIT_LEN):
            # Get current word index
            word_bin_str = (mnemonic_bin_str[i * Bip39MnemonicConst.WORD_BIT_LEN:(i + 1)
                            * Bip39MnemonicConst.WORD_BIT_LEN])
            word_idx = IntegerUtils.FromBinaryStr(word_bin_str)
            # Get word at given index
            mnemonic.append(self.m_words_list.GetWordAtIdx(word_idx))

        return Bip39Mnemonic.FromList(mnemonic)



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for Algorand mnemonic seed generation."""

# Imports
from typing import Optional, Union

from bip_utils.algorand.mnemonic.algorand_mnemonic import AlgorandLanguages
from bip_utils.algorand.mnemonic.algorand_mnemonic_decoder import AlgorandMnemonicDecoder
from bip_utils.utils.mnemonic import Mnemonic


class AlgorandSeedGenerator:
    """
    Algorand seed generator class.
    It generates the seed from a mnemonic.
    """

    m_entropy_bytes: bytes

    def __init__(self,
                 mnemonic: Union[str, Mnemonic],
                 lang: Optional[AlgorandLanguages] = AlgorandLanguages.ENGLISH) -> None:
        """
        Construct class.
        Language is set to English by default because Algorand mnemonic only support one language,
        so it's useless (and slower) to automatically detect the language.

        Args:
            mnemonic (str or Mnemonic object) : Mnemonic
            lang (AlgorandLanguages, optional): Language, None for automatic detection

        Raises:
            ValueError: If the mnemonic is not valid
        """
        self.m_entropy_bytes = AlgorandMnemonicDecoder(lang).Decode(mnemonic)

    def Generate(self) -> bytes:
        """
        Generate seed. The seed is simply the entropy bytes in Algorand case.
        There is no really need of this method, since the seed is always the same, but it's
        kept in this way to have the same usage of Bip39/Substrate seed generator
        (i.e. AlgorandSeedGenerator(mnemonic).Generate() ).

        Returns:
            bytes: Generated seed
        """
        return self.m_entropy_bytes



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module with generic coins configuration for all other modules."""

# Imports
from bip_utils.coin_conf.coin_conf import CoinConf
from bip_utils.slip.slip173 import Slip173
from bip_utils.utils.conf import CoinNames


# Bitcoin constants used by different coins
# Main net
_BTC_P2PKH_NET_VER_MN: bytes = b"\x00"
_BTC_P2SH_NET_VER_MN: bytes = b"\x05"
_BTC_P2WPKH_HRP_MN: str = Slip173.BITCOIN_MAINNET
_BTC_P2WPKH_WIT_VER_MN: int = 0
_BTC_P2TR_HRP_MN: str = Slip173.BITCOIN_MAINNET
_BTC_P2TR_WIT_VER_MN: int = 1
_BTC_WIF_NET_VER_MN: bytes = b"\x80"
# Test net
_BTC_P2PKH_NET_VER_TN: bytes = b"\x6f"
_BTC_P2SH_NET_VER_TN: bytes = b"\xc4"
_BTC_P2WPKH_HRP_TN: str = Slip173.BITCOIN_TESTNET
_BTC_P2WPKH_WIT_VER_TN: int = 0
_BTC_P2TR_HRP_TN: str = Slip173.BITCOIN_TESTNET
_BTC_P2TR_WIT_VER_TN: int = 1
_BTC_WIF_NET_VER_TN: bytes = b"\xef"


class CoinsConf:
    """Class container for coins configuration."""

    # Configuration for Acala
    Acala: CoinConf = CoinConf(
        coin_name=CoinNames("Acala", "ACA"),
        params={
            "addr_ss58_format": 10,
        },
    )

    # Configuration for Akash Network
    AkashNetwork: CoinConf = CoinConf(
        coin_name=CoinNames("Akash Network", "AKT"),
        params={
            "addr_hrp": Slip173.AKASH_NETWORK,
        },
    )

    # Configuration for Algorand
    Algorand: CoinConf = CoinConf(
        coin_name=CoinNames("Algorand", "ALGO"),
        params={},
    )

    # Configuration for Aptos
    Aptos: CoinConf = CoinConf(
        coin_name=CoinNames("Aptos", "APTOS"),
        params={
            "addr_prefix": "0x",
        },
    )

    # Configuration for Avax C-Chain
    AvaxCChain: CoinConf = CoinConf(
        coin_name=CoinNames("Avax C-Chain", "AVAX"),
        params={},
    )

    # Configuration for Avax P-Chain
    AvaxPChain: CoinConf = CoinConf(
        coin_name=CoinNames("Avax P-Chain", "AVAX"),
        params={
            "addr_hrp": "avax",
            "addr_prefix": "P-",
        },
    )

    # Configuration for Avax X-Chain
    AvaxXChain: CoinConf = CoinConf(
        coin_name=CoinNames("Avax X-Chain", "AVAX"),
        params={
            "addr_hrp": "avax",
            "addr_prefix": "X-",
        },
    )

    # Configuration for Axelar
    Axelar: CoinConf = CoinConf(
        coin_name=CoinNames("Axelar", "AXL"),
        params={
            "addr_hrp": Slip173.AXELAR,
        },
    )

    # Configuration for Band Protocol
    BandProtocol: CoinConf = CoinConf(
        coin_name=CoinNames("Band Protocol", "BAND"),
        params={
            "addr_hrp": Slip173.BAND_PROTOCOL,
        },
    )

    # Configuration for Bifrost
    Bifrost: CoinConf = CoinConf(
        coin_name=CoinNames("Bifrost", "BNC"),
        params={
            "addr_ss58_format": 6,
        },
    )

    # Configuration for Binance Chain
    BinanceChain: CoinConf = CoinConf(
        coin_name=CoinNames("Binance Chain", "BNB"),
        params={
            "addr_hrp": Slip173.BINANCE_CHAIN,
        },
    )

    # Configuration for Binance Smart Chain
    BinanceSmartChain: CoinConf = CoinConf(
        coin_name=CoinNames("Binance Smart Chain", "BNB"),
        params={},
    )

    # Configuration for Bitcoin main net
    BitcoinMainNet: CoinConf = CoinConf(
        coin_name=CoinNames("Bitcoin", "BTC"),
        params={
            "p2pkh_net_ver": _BTC_P2PKH_NET_VER_MN,
            "p2sh_net_ver": _BTC_P2SH_NET_VER_MN,
            "p2wpkh_hrp": _BTC_P2WPKH_HRP_MN,
            "p2wpkh_wit_ver": _BTC_P2WPKH_WIT_VER_MN,
            "p2tr_hrp": _BTC_P2TR_HRP_MN,
            "p2tr_wit_ver": _BTC_P2TR_WIT_VER_MN,
            "wif_net_ver": _BTC_WIF_NET_VER_MN,
        },
    )

    # Configuration for Bitcoin test net
    BitcoinTestNet: CoinConf = CoinConf(
        coin_name=CoinNames("Bitcoin TestNet", "BTC"),
        params={
            "p2pkh_net_ver": _BTC_P2PKH_NET_VER_TN,
            "p2sh_net_ver": _BTC_P2SH_NET_VER_TN,
            "p2wpkh_hrp": _BTC_P2WPKH_HRP_TN,
            "p2wpkh_wit_ver": _BTC_P2WPKH_WIT_VER_TN,
            "p2tr_hrp": _BTC_P2TR_HRP_TN,
            "p2tr_wit_ver": _BTC_P2TR_WIT_VER_MN,
            "wif_net_ver": _BTC_WIF_NET_VER_TN,
        },
    )

    # Configuration for Bitcoin Cash main net
    BitcoinCashMainNet: CoinConf = CoinConf(
        coin_name=CoinNames("Bitcoin Cash", "BCH"),
        params={
            "p2pkh_std_hrp": "bitcoincash",
            "p2pkh_std_net_ver": _BTC_P2PKH_NET_VER_MN,
            "p2pkh_legacy_net_ver": _BTC_P2PKH_NET_VER_MN,
            "p2sh_std_hrp": "bitcoincash",
            "p2sh_std_net_ver": b"\x08",
            "p2sh_legacy_net_ver": _BTC_P2SH_NET_VER_MN,
            "wif_net_ver": _BTC_WIF_NET_VER_MN,
        },
    )

    # Configuration for Bitcoin Cash test net
    BitcoinCashTestNet: CoinConf = CoinConf(
        coin_name=CoinNames("Bitcoin Cash TestNet", "BCH"),
        params={
            "p2pkh_std_hrp": "bchtest",
            "p2pkh_std_net_ver": b"\x00",
            "p2pkh_legacy_net_ver": _BTC_P2PKH_NET_VER_TN,
            "p2sh_std_hrp": "bchtest",
            "p2sh_std_net_ver": b"\x08",
            "p2sh_legacy_net_ver": _BTC_P2SH_NET_VER_TN,
            "wif_net_ver": _BTC_WIF_NET_VER_TN,
        },
    )

    # Configuration for Bitcoin Cash Simple Ledger Protocol main net
    BitcoinCashSlpMainNet: CoinConf = CoinConf(
        coin_name=CoinNames("Bitcoin Cash SLP", "SLP"),
        params={
            "p2pkh_std_hrp": "simpleledger",
            "p2pkh_std_net_ver": b"\x00",
            "p2pkh_legacy_net_ver": _BTC_P2PKH_NET_VER_MN,
            "p2sh_std_hrp": "simpleledger",
            "p2sh_std_net_ver": b"\x08",
            "p2sh_legacy_net_ver": _BTC_P2SH_NET_VER_MN,
            "wif_net_ver": _BTC_WIF_NET_VER_MN,
        },
    )

    # Configuration for Bitcoin Cash Simple Ledger Protocol test net
    BitcoinCashSlpTestNet: CoinConf = CoinConf(
        coin_name=CoinNames("Bitcoin Cash SLP TestNet", "SLP"),
        params={
            "p2pkh_std_hrp": "slptest",
            "p2pkh_std_net_ver": b"\x00",
            "p2pkh_legacy_net_ver": _BTC_P2PKH_NET_VER_TN,
            "p2sh_std_hrp": "slptest",
            "p2sh_std_net_ver": b"\x08",
            "p2sh_legacy_net_ver": _BTC_P2SH_NET_VER_TN,
            "wif_net_ver": _BTC_WIF_NET_VER_TN,
        },
    )

    # Configuration for Bitcoin SV main net
    BitcoinSvMainNet: CoinConf = CoinConf(
        coin_name=CoinNames("BitcoinSV", "BSV"),
        params={
            "p2pkh_net_ver": _BTC_P2PKH_NET_VER_MN,
            "p2sh_net_ver": _BTC_P2SH_NET_VER_MN,
            "wif_net_ver": _BTC_WIF_NET_VER_MN,
        },
    )

    # Configuration for Bitcoin SV test net
    BitcoinSvTestNet: CoinConf = CoinConf(
        coin_name=CoinNames("BitcoinSV TestNet", "BSV"),
        params={
            "p2pkh_net_ver": _BTC_P2PKH_NET_VER_TN,
            "p2sh_net_ver": _BTC_P2SH_NET_VER_TN,
            "wif_net_ver": _BTC_WIF_NET_VER_TN,
        },
    )

    # Configuration for Cardano main net
    CardanoMainNet: CoinConf = CoinConf(
        coin_name=CoinNames("Cardano", "ADA"),
        params={
            "addr_hrp": "addr",
            "staking_addr_hrp": "stake",
        },
    )

    # Configuration for Cardano test
    CardanoTestNet: CoinConf = CoinConf(
        coin_name=CoinNames("Cardano TestNet", "ADA"),
        params={
            "addr_hrp": "addr_test",
            "staking_addr_hrp": "stake_test",
        },
    )

    # Configuration for Celo
    Celo: CoinConf = CoinConf(
        coin_name=CoinNames("Celo", "CELO"),
        params={},
    )

    # Configuration for Certik
    Certik: CoinConf = CoinConf(
        coin_name=CoinNames("Certik", "CTK"),
        params={
            "addr_hrp": Slip173.CERTIK,
        },
    )

    # Configuration for ChainX
    ChainX: CoinConf = CoinConf(
        coin_name=CoinNames("ChainX", "PCX"),
        params={
            "addr_ss58_format": 44,
        },
    )

    # Configuration for Chihuahua
    Chihuahua: CoinConf = CoinConf(
        coin_name=CoinNames("Chihuahua", "HUAHUA"),
        params={
            "addr_hrp": Slip173.CHIHUAHUA,
        },
    )

    # Configuration for Cosmos
    Cosmos: CoinConf = CoinConf(
        coin_name=CoinNames("Cosmos", "ATOM"),
        params={
            "addr_hrp": Slip173.COSMOS,
        },
    )

    # Configuration for Dash main net
    DashMainNet: CoinConf = CoinConf(
        coin_name=CoinNames("Dash", "DASH"),
        params={
            "p2pkh_net_ver": b"\x4c",
            "p2sh_net_ver": b"\x10",
            "wif_net_ver": b"\xcc",
        },
    )

    # Configuration for Dash test net
    DashTestNet: CoinConf = CoinConf(
        coin_name=CoinNames("Dash TestNet", "DASH"),
        params={
            "p2pkh_net_ver": b"\x8c",
            "p2sh_net_ver": b"\x13",
            "wif_net_ver": _BTC_WIF_NET_VER_TN,
        },
    )

    # Configuration for Dogecoin main net
    DogecoinMainNet: CoinConf = CoinConf(
        coin_name=CoinNames("Dogecoin", "DOGE"),
        params={
            "p2pkh_net_ver": b"\x1e",
            "p2sh_net_ver": b"\x16",
            "wif_net_ver": b"\x9e",
        },
    )

    # Configuration for Dogecoin test net
    DogecoinTestNet: CoinConf = CoinConf(
        coin_name=CoinNames("Dogecoin TestNet", "DOGE"),
        params={
            "p2pkh_net_ver": b"\x71",
            "p2sh_net_ver": _BTC_P2SH_NET_VER_TN,
            "wif_net_ver": b"\xf1",
        },
    )

    # Configuration for eCash main net
    EcashMainNet: CoinConf = CoinConf(
        coin_name=CoinNames("eCash", "XEC"),
        params={
            "p2pkh_std_hrp": "ecash",
            "p2pkh_std_net_ver": b"\x00",
            "p2pkh_legacy_net_ver": _BTC_P2PKH_NET_VER_MN,
            "p2sh_std_hrp": "ecash",
            "p2sh_std_net_ver": b"\x08",
            "p2sh_legacy_net_ver": _BTC_P2SH_NET_VER_MN,
            "wif_net_ver": _BTC_WIF_NET_VER_MN,
        },
    )

    # Configuration for eCash test net
    EcashTestNet: CoinConf = CoinConf(
        coin_name=CoinNames("eCash TestNet", "XEC"),
        params={
            "p2pkh_std_hrp": "ectest",
            "p2pkh_std_net_ver": b"\x00",
            "p2pkh_legacy_net_ver": _BTC_P2PKH_NET_VER_TN,
            "p2sh_std_hrp": "ectest",
            "p2sh_std_net_ver": b"\x08",
            "p2sh_legacy_net_ver": _BTC_P2SH_NET_VER_TN,
            "wif_net_ver": _BTC_WIF_NET_VER_TN,
        },
    )

    # Configuration for Edgeware
    Edgeware: CoinConf = CoinConf(
        coin_name=CoinNames("Edgeware", "EDG"),
        params={
            "addr_ss58_format": 7,
        },
    )

    # Configuration for Elrond
    Elrond: CoinConf = CoinConf(
        coin_name=CoinNames("Elrond eGold", "eGLD"),
        params={
            "addr_hrp": Slip173.ELROND,
        },
    )

    # Configuration for Eos
    Eos: CoinConf = CoinConf(
        coin_name=CoinNames("EOS", "EOS"),
        params={
            "addr_prefix": "EOS",
        },
    )

    # Configuration for Ergo main net
    ErgoMainNet: CoinConf = CoinConf(
        coin_name=CoinNames("Ergo", "ERGO"),
        params={},
    )

    # Configuration for Ergo test net
    ErgoTestNet: CoinConf = CoinConf(
        coin_name=CoinNames("Ergo TestNet", "ERGO"),
        params={},
    )

    # Configuration for Ethereum
    Ethereum: CoinConf = CoinConf(
        coin_name=CoinNames("Ethereum", "ETH"),
        params={
            "addr_prefix": "0x",
        },
    )

    # Configuration for Ethereum Classic
    EthereumClassic: CoinConf = CoinConf(
        coin_name=CoinNames("Ethereum Classic", "ETC"),
        params={},
    )

    # Configuration for Fantom Opera
    FantomOpera: CoinConf = CoinConf(
        coin_name=CoinNames("Fantom Opera", "FTM"),
        params={},
    )

    # Configuration for Filecoin
    Filecoin: CoinConf = CoinConf(
        coin_name=CoinNames("Filecoin", "FIL"),
        params={
            "addr_prefix": "f",
        },
    )

    # Configuration for generic Substrate coin
    GenericSubstrate: CoinConf = CoinConf(
        coin_name=CoinNames("Generic Substrate", ""),
        params={
            "addr_ss58_format": 42,
        },
    )

    # Configuration for Harmony One
    HarmonyOne: CoinConf = CoinConf(
        coin_name=CoinNames("Harmony One", "ONE"),
        params={
            "addr_hrp": Slip173.HARMONY_ONE,
        },
    )

    # Configuration for Huobi Chain
    HuobiChain: CoinConf = CoinConf(
        coin_name=CoinNames("Huobi Token", "HT"),
        params={},
    )

    # Configuration for Icon
    Icon: CoinConf = CoinConf(
        coin_name=CoinNames("Icon", "ICX"),
        params={
            "addr_prefix": "hx",
        },
    )

    # Configuration for IRISnet
    IrisNet: CoinConf = CoinConf(
        coin_name=CoinNames("IRIS Network", "IRIS"),
        params={
            "addr_hrp": Slip173.IRIS_NETWORK,
        },
    )

    # Configuration for Karura
    Karura: CoinConf = CoinConf(
        coin_name=CoinNames("Karura", "KAR"),
        params={
            "addr_ss58_format": 8,
        },
    )

    # Configuration for Kava
    Kava: CoinConf = CoinConf(
        coin_name=CoinNames("Kava", "KAVA"),
        params={
            "addr_hrp": Slip173.KAVA,
        },
    )

    # Configuration for Kusama
    Kusama: CoinConf = CoinConf(
        coin_name=CoinNames("Kusama", "KSM"),
        params={
            "addr_ss58_format": 2,
        },
    )

    # Configuration for Litecoin main net
    LitecoinMainNet: CoinConf = CoinConf(
        coin_name=CoinNames("Litecoin", "LTC"),
        params={
            "p2pkh_std_net_ver": b"\x30",
            "p2pkh_depr_net_ver": _BTC_P2PKH_NET_VER_MN,
            "p2sh_std_net_ver": b"\x32",
            "p2sh_depr_net_ver": _BTC_P2SH_NET_VER_MN,
            "p2wpkh_hrp": Slip173.LITECOIN_MAINNET,
            "p2wpkh_wit_ver": _BTC_P2WPKH_WIT_VER_MN,
            "wif_net_ver": b"\xb0",
        },
    )

    # Configuration for Litecoin test net
    LitecoinTestNet: CoinConf = CoinConf(
        coin_name=CoinNames("Litecoin TestNet", "LTC"),
        params={
            "p2pkh_std_net_ver": b"\x6f",
            "p2pkh_depr_net_ver": _BTC_P2PKH_NET_VER_TN,
            "p2sh_std_net_ver": b"\x3a",
            "p2sh_depr_net_ver": _BTC_P2SH_NET_VER_TN,
            "p2wpkh_hrp": Slip173.LITECOIN_TESTNET,
            "p2wpkh_wit_ver": _BTC_P2WPKH_WIT_VER_TN,
            "wif_net_ver": _BTC_WIF_NET_VER_TN,
        },
    )

    # Configuration for Monero main net
    MoneroMainNet: CoinConf = CoinConf(
        coin_name=CoinNames("Monero", "XMR"),
        params={
            "addr_net_ver": b"\x12",
            "addr_int_net_ver": b"\x13",
            "subaddr_net_ver": b"\x2a",
        },
    )

    # Configuration for Monero stage net
    MoneroStageNet: CoinConf = CoinConf(
        coin_name=CoinNames("Monero StageNet", "XMR"),
        params={
            "addr_net_ver": b"\x18",
            "addr_int_net_ver": b"\x19",
            "subaddr_net_ver": b"\x24",
        },
    )

    # Configuration for Monero test net
    MoneroTestNet: CoinConf = CoinConf(
        coin_name=CoinNames("Monero TestNet", "XMR"),
        params={
            "addr_net_ver": b"\x35",
            "addr_int_net_ver": b"\x36",
            "subaddr_net_ver": b"\x3f",
        },
    )

    # Configuration for Moonbeam
    Moonbeam: CoinConf = CoinConf(
        coin_name=CoinNames("Moonbeam", "GLMR"),
        params={
            "addr_ss58_format": 1284,
        },
    )

    # Configuration for Moonriver
    Moonriver: CoinConf = CoinConf(
        coin_name=CoinNames("Moonriver", "MOVR"),
        params={
            "addr_ss58_format": 1285,
        },
    )

    # Configuration for Nano
    Nano: CoinConf = CoinConf(
        coin_name=CoinNames("Nano", "NANO"),
        params={
            "addr_prefix": "nano_",
        },
    )

    # Configuration for Near Protocol
    NearProtocol: CoinConf = CoinConf(
        coin_name=CoinNames("Near Protocol", "NEAR"),
        params={},
    )

    # Configuration for Neo
    Neo: CoinConf = CoinConf(
        coin_name=CoinNames("NEO", "NEO"),
        params={
            "addr_ver": b"\x17",
        },
    )

    # Configuration for Nine Chronicles Gold
    NineChroniclesGold: CoinConf = CoinConf(
        coin_name=CoinNames("NineChroniclesGold", "NCG"),
        params={},
    )

    # Configuration for OKEx Chain
    OkexChain: CoinConf = CoinConf(
        coin_name=CoinNames("OKExChain", "OKT"),
        params={
            "addr_hrp": Slip173.OKEX_CHAIN,
        },
    )

    # Configuration for Ontology
    Ontology: CoinConf = CoinConf(
        coin_name=CoinNames("Ontology", "ONT"),
        params={
            "addr_ver": b"\x17",
        },
    )

    # Configuration for Osmosis
    Osmosis: CoinConf = CoinConf(
        coin_name=CoinNames("Osmosis", "OSMO"),
        params={
            "addr_hrp": Slip173.OSMOSIS,
        },
    )

    # Configuration for Phala
    Phala: CoinConf = CoinConf(
        coin_name=CoinNames("Phala Network", "PHA"),
        params={
            "addr_ss58_format": 30,
        },
    )

    # Configuration for Pi Network
    PiNetwork: CoinConf = CoinConf(
        coin_name=CoinNames("Pi Network", "PI"),
        params={},
    )

    # Configuration for Plasm
    Plasm: CoinConf = CoinConf(
        coin_name=CoinNames("Plasm Network", "PLM"),
        params={
            "addr_ss58_format": 5,
        },
    )

    # Configuration for Polkadot
    Polkadot: CoinConf = CoinConf(
        coin_name=CoinNames("Polkadot", "DOT"),
        params={
            "addr_ss58_format": 0,
        },
    )

    # Configuration for Polygon
    Polygon: CoinConf = CoinConf(
        coin_name=CoinNames("Polygon", "MATIC"),
        params={},
    )

    # Configuration for Ripple
    Ripple: CoinConf = CoinConf(
        coin_name=CoinNames("Ripple", "XRP"),
        params={
            "p2pkh_net_ver": _BTC_P2PKH_NET_VER_MN,
        },
    )

    # Configuration for Secret Network
    SecretNetwork: CoinConf = CoinConf(
        coin_name=CoinNames("Secret Network", "SCRT"),
        params={
            "addr_hrp": Slip173.SECRET_NETWORK,
        },
    )

    # Configuration for Solana
    Solana: CoinConf = CoinConf(
        coin_name=CoinNames("Solana", "SOL"),
        params={},
    )

    # Configuration for Sora
    Sora: CoinConf = CoinConf(
        coin_name=CoinNames("Sora", "XOR"),
        params={
            "addr_ss58_format": 69,
        },
    )

    # Configuration for Stafi
    Stafi: CoinConf = CoinConf(
        coin_name=CoinNames("Stafi", "FIS"),
        params={
            "addr_ss58_format": 20,
        },
    )

    # Configuration for Stellar
    Stellar: CoinConf = CoinConf(
        coin_name=CoinNames("Stellar", "XLM"),
        params={},
    )

    # Configuration for Terra
    Terra: CoinConf = CoinConf(
        coin_name=CoinNames("Terra", "LUNA"),
        params={
            "addr_hrp": Slip173.TERRA,
        },
    )

    # Configuration for Tezos
    Tezos: CoinConf = CoinConf(
        coin_name=CoinNames("Tezos", "XTZ"),
        params={},
    )

    # Configuration for Theta
    Theta: CoinConf = CoinConf(
        coin_name=CoinNames("Theta Network", "THETA"),
        params={},
    )

    # Configuration for Tron
    Tron: CoinConf = CoinConf(
        coin_name=CoinNames("Tron", "TRX"),
        params={
            "addr_prefix": b"\x41",
        },
    )

    # Configuration for VeChain
    VeChain: CoinConf = CoinConf(
        coin_name=CoinNames("VeChain", "VET"),
        params={},
    )

    # Configuration for Verge
    Verge: CoinConf = CoinConf(
        coin_name=CoinNames("Verge", "XVG"),
        params={
            "p2pkh_net_ver": b"\x1e",
            "wif_net_ver": b"\x9e",
        },
    )

    # Configuration for Zcash main net
    ZcashMainNet: CoinConf = CoinConf(
        coin_name=CoinNames("Zcash", "ZEC"),
        params={
            "p2pkh_net_ver": b"\x1c\xb8",
            "p2sh_net_ver": b"\x1c\xbd",
            "wif_net_ver": _BTC_WIF_NET_VER_MN,
        },
    )

    # Configuration for Zcash test net
    ZcashTestNet: CoinConf = CoinConf(
        coin_name=CoinNames("Zcash TestNet", "ZEC"),
        params={
            "p2pkh_net_ver": b"\x1d\x25",
            "p2sh_net_ver": b"\x1c\xba",
            "wif_net_ver": _BTC_WIF_NET_VER_TN,
        },
    )

    # Configuration for Zilliqa
    Zilliqa: CoinConf = CoinConf(
        coin_name=CoinNames("Zilliqa", "ZIL"),
        params={
            "addr_hrp": Slip173.ZILLIQA,
        },
    )



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Module for SS58 decoding/encoding.
Reference: https://github.com/paritytech/substrate/wiki/External-Address-Format-(SS58)
"""

# Imports
from typing import Tuple

from bip_utils.base58 import Base58Decoder, Base58Encoder
from bip_utils.ss58.ss58_ex import SS58ChecksumError
from bip_utils.utils.crypto import Blake2b512
from bip_utils.utils.misc import BytesUtils, IntegerUtils


class SS58Const:
    """Class container for SS58 constants."""

    # Max format for simple account
    SIMPLE_ACCOUNT_FORMAT_MAX_VAL: int = 63
    # Format maximum value
    FORMAT_MAX_VAL: int = 16383
    # Reserved formats
    RESERVED_FORMATS: Tuple[int, int] = (46, 47)

    # Data length in bytes
    DATA_BYTE_LEN: int = 32

    # Checksum length in bytes
    CHECKSUM_BYTE_LEN: int = 2
    # Checksum prefix
    CHECKSUM_PREFIX: bytes = b"SS58PRE"


class _SS58Utils:
    """Class container for SS58 utility functions."""

    @staticmethod
    def ComputeChecksum(data_bytes: bytes) -> bytes:
        """
        Compute SS58 checksum.

        Args:
            data_bytes (bytes): Data bytes

        Returns:
            bytes: Computed checksum
        """
        return Blake2b512.QuickDigest(SS58Const.CHECKSUM_PREFIX + data_bytes)[:SS58Const.CHECKSUM_BYTE_LEN]


class SS58Encoder:
    """SS58 encoder class. It provides methods for encoding to SS58 format."""

    @staticmethod
    def Encode(data_bytes: bytes,
               ss58_format: int) -> str:
        """
        Encode bytes into a SS58 string.

        Args:
            data_bytes (bytes): Data bytes (32-byte length)
            ss58_format (int) : SS58 format

        Returns:
            str: SS58 encoded string

        Raises:
            ValueError: If parameters are not valid
        """

        # Check parameters
        if len(data_bytes) != SS58Const.DATA_BYTE_LEN:
            raise ValueError(f"Invalid data length ({len(data_bytes)})")
        if ss58_format < 0 or ss58_format > SS58Const.FORMAT_MAX_VAL:
            raise ValueError(f"Invalid SS58 format ({ss58_format})")
        if ss58_format in SS58Const.RESERVED_FORMATS:
            raise ValueError(f"Invalid SS58 format ({ss58_format})")

        # Simple account
        if ss58_format <= SS58Const.SIMPLE_ACCOUNT_FORMAT_MAX_VAL:
            ss58_format_bytes = IntegerUtils.ToBytes(ss58_format)
        # Full address
        else:
            # 0b00HHHHHH_MMLLLLLL -> (0b01LLLLLL, 0bHHHHHHMM)
            ss58_format_bytes = bytes([
                ((ss58_format & 0x00FC) >> 2) | 0x0040,
                (ss58_format >> 8) | ((ss58_format & 0x0003) << 6)
            ])

        # Get payload
        payload = ss58_format_bytes + data_bytes
        # Compute checksum
        checksum = _SS58Utils.ComputeChecksum(payload)
        # Encode
        return Base58Encoder.Encode(payload + checksum)


class SS58Decoder:
    """SS58 decoder class. It provides methods for decoding SS58 format."""

    @staticmethod
    def Decode(data_str: str) -> Tuple[int, bytes]:
        """
        Decode bytes from a SS58 string.

        Args:
            data_str (string): Data string

        Returns:
            tuple[int, bytes]: SS58 format and data bytes

        Raises:
            SS58ChecksumError: If checksum is not valid
            ValueError: If the string is not a valid SS58 format
        """

        # Decode string
        dec_bytes = Base58Decoder.Decode(data_str)

        # Full address
        if dec_bytes[0] & 0x40:
            ss58_format_len = 2
            ss58_format = ((dec_bytes[0] & 0x3F) << 2) | (dec_bytes[1] >> 6) | \
                          ((dec_bytes[1] & 0x3F) << 8)
        # Simple account
        else:
            ss58_format_len = 1
            ss58_format = dec_bytes[0]

        # Check format
        if ss58_format in SS58Const.RESERVED_FORMATS:
            raise ValueError(f"Invalid SS58 format ({ss58_format})")

        # Get back data and checksum
        data_bytes = dec_bytes[ss58_format_len:-SS58Const.CHECKSUM_BYTE_LEN]
        checksum_bytes = dec_bytes[-SS58Const.CHECKSUM_BYTE_LEN:]

        # Check data length
        if len(data_bytes) != SS58Const.DATA_BYTE_LEN:
            raise ValueError(f"Invalid data length ({len(data_bytes)})")

        # Compute checksum
        checksum_bytes_got = _SS58Utils.ComputeChecksum(dec_bytes[:-SS58Const.CHECKSUM_BYTE_LEN])

        # Verify checksum
        if checksum_bytes != checksum_bytes_got:
            raise SS58ChecksumError(
                f"Invalid checksum (expected {BytesUtils.ToHexString(checksum_bytes_got)}, "
                f"got {BytesUtils.ToHexString(checksum_bytes)})"
            )

        return ss58_format, data_bytes



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module containing common utility classes for mnemonic."""

# Imports
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Tuple, Type

from bip_utils.utils.misc import BytesUtils, IntegerUtils
from bip_utils.utils.mnemonic.mnemonic import Mnemonic, MnemonicLanguages
from bip_utils.utils.typing import Literal


class MnemonicUtils:
    """Class container for mnemonic utility functions."""

    @staticmethod
    def BytesChunkToWords(bytes_chunk: bytes,
                          words_list: MnemonicWordsList,
                          endianness: Literal["little", "big"]) -> List[str]:
        """
        Get words from a bytes chunk.

        Args:
            bytes_chunk (bytes)                  : Bytes chunk
            words_list (MnemonicWordsList object): Mnemonic list
            endianness ("big" or "little")       : Bytes endianness

        Returns:
            list[str]: 3 word indexes
        """
        n = words_list.Length()

        int_chunk = BytesUtils.ToInteger(bytes_chunk, endianness=endianness)

        word1_idx = int_chunk % n
        word2_idx = ((int_chunk // n) + word1_idx) % n
        word3_idx = ((int_chunk // n // n) + word2_idx) % n

        return [words_list.GetWordAtIdx(w) for w in (word1_idx, word2_idx, word3_idx)]

    @staticmethod
    def WordsToBytesChunk(word1: str,
                          word2: str,
                          word3: str,
                          words_list: MnemonicWordsList,
                          endianness: Literal["little", "big"]) -> bytes:
        """
        Get bytes chunk from words.

        Args:
            word1 (str)                          : Word 1
            word2 (str)                          : Word 2
            word3 (str)                          : Word 3
            words_list (MnemonicWordsList object): Mnemonic list
            endianness ("big" or "little")       : Bytes endianness

        Returns:
            bytes: Bytes chunk
        """
        n = words_list.Length()

        # Get the word indexes
        word1_idx = words_list.GetWordIdx(word1)
        word2_idx = words_list.GetWordIdx(word2) % n
        word3_idx = words_list.GetWordIdx(word3) % n

        # Get back the bytes chunk
        int_chunk = word1_idx + (n * ((word2_idx - word1_idx) % n)) + (n * n * ((word3_idx - word2_idx) % n))

        return IntegerUtils.ToBytes(int_chunk, bytes_num=4, endianness=endianness)


class MnemonicWordsList:
    """Mnemonic words list class."""

    m_idx_to_words: List[str]
    m_words_to_idx: Dict[str, int]

    def __init__(self,
                 words_list: List[str]) -> None:
        """
        Construct class.

        Args:
            words_list (list[str]): Words list
        """
        self.m_idx_to_words = words_list
        # Map strings to indexes as well for a quick word searching
        self.m_words_to_idx = {words_list[i]: i for i in range(len(words_list))}

    def Length(self) -> int:
        """
        Get the length of the words list.

        Returns:
            int: Words list length
        """
        return len(self.m_idx_to_words)

    def GetWordIdx(self,
                   word: str) -> int:
        """
        Get the index of the specified word.

        Args:
            word (str): Word to be searched

        Returns:
            int: Word index

        Raises:
            ValueError: If the word is not found
        """
        try:
            return self.m_words_to_idx[word]
        except KeyError as ex:
            raise ValueError(f"Unable to find word {word}") from ex

    def GetWordAtIdx(self,
                     word_idx: int) -> str:
        """
        Get the word at the specified index.

        Args:
            word_idx (int): Word index

        Returns:
            str: Word at the specified index
        """
        return self.m_idx_to_words[word_idx]


class MnemonicWordsListFileReader:
    """
    Mnemonic words list file reader class.
    It reads the words list from a file.
    """

    @staticmethod
    def LoadFile(file_path: str,
                 words_num: int) -> MnemonicWordsList:
        """
        Load words list file correspondent to the specified language.

        Args:
            file_path (str): File name
            words_num (int): Number of expected words

        Returns:
            MnemonicWordsList: MnemonicWordsList object

        Raises:
            ValueError: If loaded words list is not valid
        """

        # Read file
        with open(file_path, "r", encoding="utf-8") as fin:
            words_list = [word.strip()
                          for word in fin.readlines()
                          if word.strip() != "" and not word.startswith("#")]

        # Check words list count
        if len(words_list) != words_num:
            raise ValueError(f"Number of loaded words list ({len(words_list)}) is not valid")

        return MnemonicWordsList(words_list)


class MnemonicWordsListGetterBase(ABC):
    """Mnemonic words list getter base class."""

    m_words_lists: Dict[MnemonicLanguages, MnemonicWordsList]

    # Global instance
    __instance: Optional[MnemonicWordsListGetterBase] = None

    def __init__(self):
        """Construct class."""
        self.m_words_lists = {}

    @abstractmethod
    def GetByLanguage(self,
                      lang: MnemonicLanguages) -> MnemonicWordsList:
        """
        Get words list by language.
        Words list of a specific language are loaded from file only the first time they are requested.

        Args:
            lang (MnemonicLanguages): Language

        Returns:
            MnemonicWordsList object: MnemonicWordsList object

        Raises:
            TypeError: If the language is not of the correct enumerative
            ValueError: If loaded words list is not valid
        """

    def _LoadWordsList(self,
                       lang: MnemonicLanguages,
                       file_name: str,
                       words_num: int) -> MnemonicWordsList:
        """
        Load words list.

        Args:
            lang (MnemonicLanguages): Language
            file_name (str)         : File name
            words_num (int)         : Number of expected words

        Returns:
            MnemonicWordsList object: MnemonicWordsList object

        Raises:
            ValueError: If loaded words list is not valid
        """

        # Only load words list for a specific language the first time it is requested
        try:
            return self.m_words_lists[lang]
        except KeyError:
            self.m_words_lists[lang] = MnemonicWordsListFileReader.LoadFile(file_name,
                                                                            words_num)

            return self.m_words_lists[lang]

    @classmethod
    def Instance(cls) -> MnemonicWordsListGetterBase:
        """
        Get the global class instance.

        Returns:
            MnemonicWordsListGetterBase object: MnemonicWordsListGetterBase object
        """
        if cls.__instance is None:
            cls.__instance = cls()
        return cls.__instance


class MnemonicWordsListFinderBase(ABC):
    """
    Mnemonic words list finder base class.
    It automatically finds the correct words list from a mnemonic.
    """

    @classmethod
    @abstractmethod
    def FindLanguage(cls,
                     mnemonic: Mnemonic) -> Tuple[MnemonicWordsList, MnemonicLanguages]:
        """
        Automatically find the language of the specified mnemonic and get the correct MnemonicWordsList class for it.

        Args:
            mnemonic (Mnemonic object): Mnemonic object

        Returns:
           tuple[MnemonicWordsList, MnemonicLanguages]: MnemonicWordsList object (index 0), mnemonic language (index 1)

        Raises:
            ValueError: If the mnemonic language cannot be found
        """

    @staticmethod
    def _FindLanguageGeneric(
            mnemonic: Mnemonic,
            langs_enum: Type[MnemonicLanguages],
            words_list_getter_cls: Type[MnemonicWordsListGetterBase]
    ) -> Tuple[MnemonicWordsList, MnemonicLanguages]:
        """
        Automatically find the language of the specified mnemonic and get the correct MnemonicWordsList class for it.
        Generic version that doesn't depending on a specific mnemonic type.

        Args:
            mnemonic (Mnemonic object)                         : Mnemonic object
            langs_enum (MnemonicLanguages class)               : Language class
            words_list_getter_cls (MnemonicWordsListGetterBase): Word list getter class type

        Returns:
           tuple[MnemonicWordsList, MnemonicLanguages]: MnemonicWordsList object (index 0), mnemonic language (index 1)

        Raises:
            ValueError: If the mnemonic language cannot be found
        """
        for lang in langs_enum:
            # Search all the words because some languages have words in common
            # (e.g. 'fatigue' both in English and French)
            # It's more time consuming, but considering only the first word can detect the wrong language sometimes
            try:
                words_list = words_list_getter_cls.Instance().GetByLanguage(lang)
                for word in mnemonic.ToList():
                    words_list.GetWordIdx(word)
                return words_list, lang
            except ValueError:
                continue

        # Language not found
        raise ValueError(f"Invalid language for mnemonic '{mnemonic.ToStr()}'")



from bip_utils.substrate.conf.substrate_coin_conf import SubstrateCoinConf
from bip_utils.substrate.conf.substrate_coins import SubstrateCoins
from bip_utils.substrate.conf.substrate_conf import SubstrateConf
from bip_utils.substrate.conf.substrate_conf_getter import SubstrateConfGetter



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for CIP-1852 coins configuration."""

# Imports
from bip_utils.addr import AdaShelleyAddrEncoder, AdaShelleyAddrNetworkTags
from bip_utils.bip.bip32 import Bip32Const, Bip32KholawEd25519
from bip_utils.bip.conf.common import DER_PATH_NON_HARDENED_FULL, BipCoinConf
from bip_utils.cardano.bip32.cardano_icarus_bip32 import CardanoIcarusBip32
from bip_utils.coin_conf import CoinsConf
from bip_utils.slip.slip44 import Slip44


class Cip1852Conf:
    """Class container for CIP-1852 configuration."""

    # Configuration for Cardano main net (Icarus)
    CardanoIcarusMainNet: BipCoinConf = BipCoinConf(
        coin_names=CoinsConf.CardanoMainNet.CoinNames(),
        coin_idx=Slip44.CARDANO,
        is_testnet=False,
        def_path=DER_PATH_NON_HARDENED_FULL,
        key_net_ver=Bip32Const.KHOLAW_KEY_NET_VERSIONS,
        wif_net_ver=None,
        bip32_cls=CardanoIcarusBip32,
        addr_cls=AdaShelleyAddrEncoder,
        addr_params={
            "net_tag": AdaShelleyAddrNetworkTags.MAINNET,
        },
    )

    # Configuration for Cardano test net (Icarus)
    CardanoIcarusTestNet: BipCoinConf = BipCoinConf(
        coin_names=CoinsConf.CardanoTestNet.CoinNames(),
        coin_idx=Slip44.CARDANO,
        is_testnet=True,
        def_path=DER_PATH_NON_HARDENED_FULL,
        key_net_ver=Bip32Const.TEST_NET_KEY_NET_VERSIONS,
        wif_net_ver=None,
        bip32_cls=CardanoIcarusBip32,
        addr_cls=AdaShelleyAddrEncoder,
        addr_params={
            "net_tag": AdaShelleyAddrNetworkTags.TESTNET,
        },
    )

    # Configuration for Cardano main net (Ledger)
    CardanoLedgerMainNet: BipCoinConf = BipCoinConf(
        coin_names=CoinsConf.CardanoMainNet.CoinNames(),
        coin_idx=Slip44.CARDANO,
        is_testnet=False,
        def_path=DER_PATH_NON_HARDENED_FULL,
        key_net_ver=Bip32Const.KHOLAW_KEY_NET_VERSIONS,
        wif_net_ver=None,
        bip32_cls=Bip32KholawEd25519,
        addr_cls=AdaShelleyAddrEncoder,
        addr_params={
            "net_tag": AdaShelleyAddrNetworkTags.MAINNET,
        },
    )

    # Configuration for Cardano test net (Ledger)
    CardanoLedgerTestNet: BipCoinConf = BipCoinConf(
        coin_names=CoinsConf.CardanoTestNet.CoinNames(),
        coin_idx=Slip44.CARDANO,
        is_testnet=True,
        def_path=DER_PATH_NON_HARDENED_FULL,
        key_net_ver=Bip32Const.TEST_NET_KEY_NET_VERSIONS,
        wif_net_ver=None,
        bip32_cls=Bip32KholawEd25519,
        addr_cls=AdaShelleyAddrEncoder,
        addr_params={
            "net_tag": AdaShelleyAddrNetworkTags.TESTNET,
        },
    )



# Copyright (c) 2022 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for Electrum v2 mnemonic seed generation."""

# Imports
from typing import Optional, Union

from bip_utils.electrum.mnemonic_v2.electrum_v2_mnemonic import ElectrumV2Languages, ElectrumV2Mnemonic
from bip_utils.electrum.mnemonic_v2.electrum_v2_mnemonic_validator import ElectrumV2MnemonicValidator
from bip_utils.utils.crypto import Pbkdf2HmacSha512
from bip_utils.utils.misc import StringUtils
from bip_utils.utils.mnemonic import Mnemonic


class ElectrumV2SeedGeneratorConst:
    """Class container for Electrum seed generator constants (v2)."""

    # Salt modifier for seed generation
    SEED_SALT_MOD: str = "electrum"
    # PBKDF2 round for seed generation
    SEED_PBKDF2_ROUNDS: int = 2048


class ElectrumV2SeedGenerator:
    """
    Electrum seed generator class (v2).
    It generates the seed from a mnemonic.
    """

    m_entropy_bytes: bytes

    def __init__(self,
                 mnemonic: Union[str, Mnemonic],
                 lang: Optional[ElectrumV2Languages] = None) -> None:
        """
        Construct class.

        Args:
            mnemonic (str or Mnemonic object)   : Mnemonic
            lang (ElectrumV2Languages, optional): Language, None for automatic detection

        Raises:
            ValueError: If the mnemonic is not valid
        """
        ElectrumV2MnemonicValidator(lang=lang).Validate(mnemonic)
        self.m_mnemonic = (ElectrumV2Mnemonic.FromString(mnemonic)
                           if isinstance(mnemonic, str)
                           else mnemonic)

    def Generate(self,
                 passphrase: str = "") -> bytes:
        """
        Generate the seed using the specified passphrase.

        Args:
            passphrase (str, optional): Passphrase, empty if not specified

        Returns:
            bytes: Generated seed
        """
        salt = StringUtils.NormalizeNfkd(ElectrumV2SeedGeneratorConst.SEED_SALT_MOD + passphrase)
        return Pbkdf2HmacSha512.DeriveKey(self.m_mnemonic.ToStr(),
                                          salt,
                                          ElectrumV2SeedGeneratorConst.SEED_PBKDF2_ROUNDS)
